"""
Bar Indicator
-------------
"""

from qtpyvcp.widgets import VCPWidget
from qtpyvcp.widgets.base_widgets.bar_indicator import BarIndicatorBase

from qtpyvcp.utilities import logger
LOG = logger.getLogger(__name__)


class BarIndicator(BarIndicatorBase, VCPWidget):
    """General purpose bars style indicator.

    Args:
        parent (QWidget) : The parent widget, or None.
    """

    DEFAULT_RULE_PROPERTY = 'Value'
    RULE_PROPERTIES = VCPWidget.RULE_PROPERTIES.copy()
    RULE_PROPERTIES.update({
        'Value': ['setValue', int],
    })

    def __init__(self, parent=None):
        super(BarIndicator, self).__init__(parent)



#!/usr/bin/env python

#   Copyright (c) 2018 Kurt Jacobson
#      <kurtcjacobson@gmail.com>
#
#   This file is part of QtPyVCP.
#
#   QtPyVCP is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 2 of the License, or
#   (at your option) any later version.
#
#   QtPyVCP is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with QtPyVCP.  If not, see <http://www.gnu.org/licenses/>.

import sys

from datetime import datetime

from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QWidget, QGridLayout, QLabel, QPushButton, QDesktopWidget, QVBoxLayout, QApplication, QDialog


class Message(QWidget):
    def __init__(self, title, message, parent=None):
        super(Message, self).__init__(parent)
        self.setLayout(QGridLayout())
        self.titleLabel = QLabel(title, self)
        self.titleLabel.setStyleSheet(
            """
            font-family: 'Roboto', sans-serif;
            font-size: 14px; font-weight: bold;
            padding: 0;
            """)
        self.messageLabel = QLabel(message, self)
        self.messageLabel.setStyleSheet(
            """
            font-family: 'Roboto', sans-serif;
            font-size: 12px;
            font-weight: normal;
            padding: 0;
            """)
        self.buttonClose = QPushButton(self)
        self.buttonClose.setIcon(QIcon.fromTheme("window-close"))
        self.buttonClose.setFixedSize(32, 32)
        self.layout().addWidget(self.titleLabel, 0, 0)
        self.layout().addWidget(self.messageLabel, 1, 0)
        self.layout().addWidget(self.buttonClose, 0, 1, 2, 1)


class NativeNotification(QDialog):
    signNotifyClose = Signal(str)

    def __init__(self, parent=None):
        super(NativeNotification, self).__init__(parent=parent)
        time = datetime.now()

        current_time = "{}:{}".format(time.hour, time.minute)

        self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint | Qt.WindowSystemMenuHint)

        resolution = QDesktopWidget().screenGeometry(-1)
        screenWidth = resolution.width()
        screenHeight = resolution.height()

        self.maxMessages = 5
        self.nMessages = 0
        self.activeMessages = list()
        self.mainLayout = QVBoxLayout(self)
        self.move(screenWidth, 0)

    def setNotify(self, title, message):
        m = Message(title, message, self)
        self.mainLayout.insertWidget(0, m)
        m.buttonClose.clicked.connect(self.onClicked)
        self.nMessages += 1

        if self.nMessages > self.maxMessages:
            prev_message = self.activeMessages.pop(0)
            self.mainLayout.removeWidget(prev_message)
            prev_message.deleteLater()
            self.nMessages -= 1

        self.activeMessages.append(m)

        self.show()
        self.raise_()

    def onClicked(self):
        m = self.sender().parent()
        self.mainLayout.removeWidget(m)
        self.activeMessages.remove(m)
        m.deleteLater()
        self.nMessages -= 1
        self.adjustSize()
        if self.nMessages == 0:
            self.close()

    def popNotify(self):
        """Removes the last notification in the list"""
        pass


class Example(QWidget):
    counter = 0

    def __init__(self, parent=None):
        super(Example, self).__init__(parent)
        self.setLayout(QVBoxLayout())
        btn = QPushButton("Send Notify", self)
        self.layout().addWidget(btn)

        self.notification = NativeNotification()
        btn.clicked.connect(self.notify)
        self.index = 0

    def notify(self):
        self.notification.setNotify("BroadCast System",
                                    "This is a test of the broadcast system.{}".format(self.index))
        self.index += 1


if __name__ == '__main__':
    app = QApplication(sys.argv)
    w = Example()
    w.show()
    sys.exit(app.exec_())




from qtpy.QtWidgets import QGroupBox

from qtpyvcp import hal
from qtpyvcp.widgets import HALWidget


class HalGroupBox(QGroupBox, HALWidget):
    """HAL GroupBox

    GroupBox that can be enabled/disabled via HAL pins.

    .. table:: Generated HAL Pins

        ========================= ===== =========
        HAL Pin Name              Type  Direction
        ========================= ===== =========
        qtpyvcp.group-box.enable   bit   in
        qtpyvcp.group-box.visible  bit   in
        qtpyvcp.group-box.check    bit   in
        qtpyvcp.group-box.checked  bit   out
        ========================= ===== =========
    """

    def __init__(self, parent=None):
        super(HalGroupBox, self).__init__(parent)

        self._enable_pin = None
        self._visible_pin = None
        self._check_pin = None
        self._checked_pin = None

        self.toggled.connect(self.onCheckedStateChanged)

    def onCheckedStateChanged(self, checked):
        if self._checked_pin is not None:
            self._checked_pin.value = checked

    def initialize(self):
        comp = hal.getComponent()
        obj_name = self.getPinBaseName()

        # add group-box.enable HAL pin
        self._enable_pin = comp.addPin(obj_name + ".enable", "bit", "in")
        self._enable_pin.value = self.isEnabled()
        self._enable_pin.valueChanged.connect(self.setEnabled)

        # add group-box.visible HAL pin
        self._visible_pin = comp.addPin(obj_name + ".visible", "bit", "in")
        self._visible_pin.value = self.isVisible()
        self._visible_pin.valueChanged.connect(self.setVisible)

        if self.isCheckable():
            # add group-box.check HAL pin
            self._check_pin = comp.addPin(obj_name + ".check", "bit", "in")
            self._check_pin.value = self.isChecked()
            self._check_pin.valueChanged.connect(self.setChecked)

            # add group-box.checked HAL pin
            self._checked_pin = comp.addPin(obj_name + ".checked", "bit", "out")
            self._checked_pin.value = self.isChecked()



import linuxcnc

# Set up logging
from qtpyvcp.utilities import logger
LOG = logger.getLogger(__name__)

from qtpyvcp.plugins import getPlugin

STATUS = getPlugin('status')
STAT = STATUS.stat

CMD = linuxcnc.command()

#==============================================================================
# Coolent actions
#==============================================================================

class flood:
    """Flood Actions Group"""
    @staticmethod
    def on():
        """Turns Flood coolant ON

        ActionButton syntax::

            coolant.flood.on
        """
        LOG.debug("Turning Flood coolant green<ON>")
        CMD.flood(linuxcnc.FLOOD_ON)

    @staticmethod
    def off():
        """Turns Flood coolant OFF

        ActionButton syntax::

            coolant.flood.off
        """
        LOG.debug("Turning Flood coolant red<OFF>")
        CMD.flood(linuxcnc.FLOOD_OFF)

    @staticmethod
    def toggle():
        """Toggles Flood coolant ON/OFF

        ActionButton syntax::

            coolant.flood.toggle
        """
        if STAT.flood == linuxcnc.FLOOD_ON:
            flood.off()
        else:
            flood.on()

class mist:
    """Mist Actions Group"""
    @staticmethod
    def on():
        """Turns Mist coolant ON

        ActionButton syntax::

            coolant.mist.on
        """
        LOG.debug("Turning Mist coolant green<ON>")
        CMD.mist(linuxcnc.MIST_ON)

    @staticmethod
    def off():
        """Turns Mist coolant OFF

        ActionButton syntax::

            coolant.mist.off
        """
        LOG.debug("Turning Mist coolant red<OFF>")
        CMD.mist(linuxcnc.MIST_OFF)

    @staticmethod
    def toggle():
        """Toggles Mist coolant ON/OFF

        ActionButton syntax::

            coolant.mist.toggle
        """
        if STAT.mist == linuxcnc.MIST_ON:
            mist.off()
        else:
            mist.on()

def _coolant_ok(widget=None):
    """Checks if it is OK to turn coolant ON.

    Args:
        widget (QWidget, optional) : Widget to enable/disable according to result.

    Atributes:
        msg (string) : The reason the action is not permitted, or empty if permitted.

    Retuns:
        bool : True if OK, else False.
    """
    if STAT.task_state == linuxcnc.STATE_ON:
        ok = True
        msg = ""
    else:
        ok = False
        msg = "Can't turn on coolant when machine is not ON"

    _coolant_ok.msg = msg

    if widget is not None:
        widget.setEnabled(ok)
        widget.setStatusTip(msg)
        widget.setToolTip(msg)

    return ok

def _flood_bindOk(widget):
    _coolant_ok(widget)
    widget.setChecked(STAT.flood == linuxcnc.FLOOD_ON)
    STATUS.task_state.onValueChanged(lambda: _coolant_ok(widget))
    STATUS.flood.onValueChanged(lambda s: widget.setChecked(s == linuxcnc.FLOOD_ON))

def _mist_bindOk(widget):
    _coolant_ok(widget)
    widget.setChecked(STAT.mist == linuxcnc.MIST_ON)
    STATUS.task_state.onValueChanged(lambda: _coolant_ok(widget))
    STATUS.mist.onValueChanged(lambda s: widget.setChecked(s == linuxcnc.MIST_ON))

flood.on.ok = flood.off.ok = flood.toggle.ok = _coolant_ok
flood.on.bindOk = flood.off.bindOk = flood.toggle.bindOk = _flood_bindOk

mist.on.ok = mist.off.ok = mist.toggle.ok = _coolant_ok
mist.on.bindOk = mist.off.bindOk = mist.toggle.bindOk = _mist_bindOk





from qtpy.QtWidgets import QDoubleSpinBox
from qtpy.QtCore import QEvent

from qtpyvcp import hal
from qtpyvcp.widgets import HALWidget


class HalDoubleSpinBox(QDoubleSpinBox, HALWidget):
    """HAL DoubleSpinBox

    DoubleSpinBox for displaying and setting `float` HAL pin values.

    .. table:: Generated HAL Pins

        ========================= ========= =========
        HAL Pin Name              Type      Direction
        ========================= ========= =========
        qtpyvcp.spinbox.enable    bit       in
        qtpyvcp.spinbox.in        float     in
        qtpyvcp.spinbox.out       float     out
        ========================= ========= =========
    """
    def __init__(self, parent=None):
        super(HalDoubleSpinBox, self).__init__(parent)

        self._value_pin = None
        self._enabled_pin = None

        self._signed_int = True

        self.valueChanged.connect(self.onCheckedStateChanged)

    def changeEvent(self, event):
        super(HalDoubleSpinBox, self).changeEvent(event)
        if event == QEvent.EnabledChange and self._enabled_pin is not None:
            self._enabled_pin.value = self.isEnabled()

    def onCheckedStateChanged(self, checked):
        if self._value_pin is not None:
            self._value_pin.value = checked

    def initialize(self):
        comp = hal.getComponent()
        obj_name = self.getPinBaseName()

        # add spinbox.enabled HAL pin
        self._enabled_pin = comp.addPin(obj_name + ".enable", "bit", "in")
        self._enabled_pin.value = self.isEnabled()
        self._enabled_pin.valueChanged.connect(self.setEnabled)

        # add spinbox.checked HAL pin
        self._value_pin = comp.addPin(obj_name + ".out", "float", "out")
        self._value_pin.value = self.value()

        # add spinbox.checked HAL pin
        self._set_value_pin = comp.addPin(obj_name + ".in", "float", "in")
        self._set_value_pin.valueChanged.connect(self.setValue)



from qtpyvcp.widgets.qtdesigner import _DesignerPlugin, _PluginExtension

from frame import VCPFrame
class VCPFramePlugin(_DesignerPlugin):
    def pluginClass(self):
        return VCPFrame
    def isContainer(self):
        return True

from widget import VCPWidget
class VCPWidgetPlugin(_DesignerPlugin):
    def pluginClass(self):
        return VCPWidget
    def isContainer(self):
        return True

from stack import VCPStackedWidget
class VCPStackedWidgetPlugin(_DesignerPlugin):
    def pluginClass(self):
        return VCPStackedWidget
    def isContainer(self):
        return True
    def domXml(self):
        return """
    <widget class="VCPStackedWidget" name="stackedWidget">
    <property name="geometry">
     <rect>
      <x>120</x>
      <y>420</y>
      <width>371</width>
      <height>221</height>
     </rect>
    </property>
    <widget class="QWidget" name="page_1"/>
    <widget class="QWidget" name="page_2"/>
   </widget>"""



"""Smart Axis and Join positions plugin.

Calculates Absolute, Relative, and Distance-To-Go values for each axis.
Configurable to report actual or commanded positions and values in either
machine or program units.

Usage:
    Example syntax to use in Widget Rules Editor::

        position:abs?string&axis=x        # returns X axis absolute position
        position:rel?string&axis=x        # returns X axis relative position
        position:dtg?string&axis=x        # returns X axis DTG value


YAML configuration:

This goes in your config.yml file in the `data_plugins` section.

.. code-block:: yaml

    data_plugins:
      position:
        kwargs:
          # whether to report actual or commanded pos
          report_actual_pos: False
          # whether to report program or machine units
          use_program_units: True
          # format used for metric units
          metric_format: "%9.3f"
          # format used for imperial units
          imperial_format: "%8.4f"

ToDO:
    Add joint positions.
"""

import math

from qtpyvcp.utilities.info import Info
from qtpyvcp.utilities.logger import getLogger
from qtpyvcp.plugins import DataPlugin, DataChannel, getPlugin

STATUS = getPlugin('status')
STAT = STATUS.stat
INFO = Info()

# Set up logging
LOG = getLogger(__name__)

MACHINE_COORDS = INFO.getCoordinates()
MACHINE_UNITS = 2 if INFO.getIsMachineMetric() else 1

# Set the conversions used for changing the DRO units
# Only convert linear axes (XYZUVW), use factor of unity for ABC
if MACHINE_UNITS == 2:
    # List of factors for converting from mm to inches
    CONVERSION_FACTORS = [1.0 / 25.4] * 3 + [1] * 3 + [1.0 / 25.4] * 3
else:
    # List of factors for converting from inches to mm
    CONVERSION_FACTORS = [25.4] * 3 + [1] * 3 + [25.4] * 3


class Position(DataPlugin):
    """Positions Plugin"""
    def __init__(self, report_actual_pos=False, use_program_units=True,
                 metric_format='%9.3f', imperial_format='%8.4f'):
        super(Position, self).__init__()

        self._report_actual_pos = False
        self._use_program_units = use_program_units
        self._metric_format = metric_format
        self._imperial_format = imperial_format

        self._current_format = self._imperial_format

        self._update()

        # all these should cause the positions to update
        STATUS.position.signal.connect(self._update)
        STATUS.g5x_offset.signal.connect(self._update)
        STATUS.g92_offset.signal.connect(self._update)
        STATUS.tool_offset.signal.connect(self._update)
        STATUS.program_units.signal.connect(self.updateUnits)

        self.report_actual_pos = report_actual_pos

    def getChannel(self, url):
        """Get data channel from URL.

        Args:
            url (str) : The URL of the channel to get.

        Returns:
            tuple : (chan_obj, chan_exp)
        """

        chan, sep, query = url.partition('?')
        raw_args = query.split('&')

        args = []
        kwargs = {}
        for arg in [a for a in raw_args if a != '']:
            if '=' in arg:
                key, val = arg.split('=')
                kwargs[key] = val
            else:
                args.append(arg)

        try:
            chan_obj = self.channels[chan]

            if 'axis' in kwargs:
                axis = kwargs.pop('axis')
                try:
                    kwargs['anum'] = int(axis)
                except ValueError:
                    kwargs['anum'] = 'xyzabcuvw'.index(str(axis).lower())

            if len(args) > 0 and args[0] in ('string', 'text', 'str'):
                chan_exp = lambda: chan_obj.getString(*args[1:], **kwargs)
            else:
                chan_exp = lambda: chan_obj.getValue(*args, **kwargs)

        except (KeyError, SyntaxError):
            LOG.exception('Error getting channel')
            return None, None

        return chan_obj, chan_exp

    def updateUnits(self, canon_units):
        print 'updating units', canon_units
        if canon_units == 2:
            self._current_format = self._metric_format
        else:
            self._current_format = self._imperial_format

        self._update()

    @DataChannel
    def rel(self, chan, anum=-1):
        """The current relative axis positions including all offsets

        To get a single axis pass string and the axis letter::

            position:rel?string&axis=x

        To get a tuple of all the axes pass only string::

            position:rel?

        :returns: current relative axis positions including all offsets
        :rtype: tuple, str
        """

        if anum == -1:
            return chan.value
        return chan.value[anum]

    @rel.tostring
    def rel(self, chan, anum):
        return self._current_format % chan.value[anum]

    @DataChannel
    def abs(self, chan, anum=-1):
        """The current absolute axis positions

        To get a single axis pass string and the axis letter::

            position:abs?string&axis=x

        To get a tuple of all the axes pass only string::

            position:abs?

        :returns: current absolute axis positions
        :rtype: tuple, str
        """

        if anum == -1:
            return chan.value
        return chan.value[anum]

    @abs.tostring
    def abs(self, chan, anum):
        return self._current_format % chan.value[anum]


    @DataChannel
    def dtg(self, chan, anum=-1):
        """The remaining distance-to-go for the current move

        To get a single axis pass string and the axis letter::

            position:dtg?string&axis=x

        To get a tuple of all the axes pass only string::

            position:dtg?

        :returns: remaining distance-to-go for the current move
        :rtype: tuple, str
        """

        if anum == -1:
            return chan.value
        return chan.value[anum]

    @dtg.tostring
    def dtg(self, chan, anum):
        return self._current_format % chan.value[anum]

    @property
    def report_actual_pos(self):
        """Whether to report the actual position. Default True.

        See the YAML configuration.

        """
        return self._report_actual_pos

    @report_actual_pos.setter
    def report_actual_pos(self, report_actual_pos):
        if report_actual_pos == self._report_actual_pos:
            return
        self._report_actual_pos = report_actual_pos

        if self._report_actual_pos:
            # disconnect commanded pos update signals
            STATUS.position.signal.disconnect(self._update)
            # STATUS.joint_position.signal.disconnect(self._update)
            # connect actual pos update signals
            STATUS.actual_position.signal.connect(self._update)
            # STATUS.joint_actual_position.signal.connect(self.joint._update)
        else:
            # disconnect actual pos update signals
            STATUS.actual_position.signal.disconnect(self._update)
            # STATUS.joint_actual_position.signal.disconnect(self._update)
            # connect commanded pos update signals
            STATUS.position.signal.connect(self._update)
            # STATUS.joint_position.signal.connect(self._update)

    def _update(self):

        if self._report_actual_pos:
            pos = STAT.actual_position
        else:
            pos = STAT.position

        dtg = STAT.dtg
        g5x_offset = STAT.g5x_offset
        g92_offset = STAT.g92_offset
        tool_offset = STAT.tool_offset

        rel = [0] * 9
        for axis in INFO.AXIS_NUMBER_LIST:
            rel[axis] = pos[axis] - g5x_offset[axis] - tool_offset[axis]

        if STAT.rotation_xy != 0:
            t = math.radians(-STAT.rotation_xy)
            xr = rel[0] * math.cos(t) - rel[1] * math.sin(t)
            yr = rel[0] * math.sin(t) + rel[1] * math.cos(t)
            rel[0] = xr
            rel[1] = yr

        for axis in INFO.AXIS_NUMBER_LIST:
            rel[axis] -= g92_offset[axis]

        if STAT.program_units != MACHINE_UNITS and self._use_program_units:
            pos = [pos[anum] * CONVERSION_FACTORS[anum] for anum in range(9)]
            rel = [rel[anum] * CONVERSION_FACTORS[anum] for anum in range(9)]
            dtg = [dtg[anum] * CONVERSION_FACTORS[anum] for anum in range(9)]

        self.rel.setValue(tuple(rel))
        self.abs.setValue(tuple(pos))
        self.dtg.setValue(tuple(dtg))



#!/usr/bin/env python

from qtpy.QtCore import Slot
from qtpyvcp.widgets.form_widgets.main_window import VCPMainWindow

# Setup logging
from qtpyvcp.utilities import logger
LOG = logger.getLogger('qtpyvcp.' + __name__)

from qtpyvcp import actions

class MainWindow(VCPMainWindow):
    def __init__(self, *args, **kwargs):
        super(MainWindow, self).__init__(*args, **kwargs)

        self.setWindowTitle("Brender VCP")

    # ==========================================================================
    #  Add/Override methods and slots below to customize the main window
    # ==========================================================================

        actions.bindWidget(self.flood, action='coolant.flood.toggle')
        actions.bindWidget(self.floodCheckBox, action="coolant.flood.toggle")

    # This slot will be automatically connected to a menu item named 'Test'
    # created in QtDesigner.
    @Slot()
    def on_actionTest_triggered(self):
        print 'Test action triggered'
        # implement the action here




import os
import ast
import ConfigParser

from qtpyvcp.utilities.logger import getLogger
log = getLogger(__name__)


class RuntimeConfig(object):
    def __init__(self, fname='~/.qtpyvcprc'):
        self.fname = fname
        self._cp = ConfigParser.RawConfigParser()
        self._cp.optionxform = str

        self.require_write = False

        self.getters = {
            bool: self.getBool,
            float: self.getFloat,
            int: self.getInt,
            list: self.getList,
            dict: self.getDict,
            str: self.getStr,
        }

    def read(self, fname=None):
        self.fname = os.path.abspath(os.path.expanduser(fname or self.fname))
        self._cp.read(self.fname)
        return self

    def write(self, fname=None):
        fname = fname or self.fname
        with open(fname, 'w') as fh:
            self._cp.write(fh)

    def get(self, section, option, default=None):
        try:
            typ = (type(default) if default is not None else str)
            getter = self.getters.get(typ)
            value = getter(section, option, default)
            return value
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            # Add the section and the option
            self.set(section, option, default)

        return default

    def set(self, section, option, value):
        try:
            self._cp.set(section, option, str(value))
        except ConfigParser.NoSectionError:
            # Add the section and the option
            self._cp.add_section(section)
            self._cp.set(section, option, str(value))

        self.require_write = True

    def getStr(self, section, option, default):
        return self._cp.get(section, option)

    def getFloat(self, section, option, default):
        value = self._cp.get(section, option)
        try:
            return float(value)
        except (ValueError, SyntaxError):
            return default

    def getInt(self, section, option, default):
        value = self._cp.get(section, option)
        try:
            return int(value)
        except (ValueError, SyntaxError):
            return default

    def getBool(self, section, option, default):
        value = self._cp.get(section, option)
        if value.lower() in ['1', 'true', 'yes', 'on']:
            return True
        elif value.lower() in ['0', 'false', 'no', 'off']:
            return False
        log.error('Non boolean value "{0}" for option [{1}] {2},'
            ' using default value of "{3}"'.format(value, section, option, default))
        return default

    def getList(self, section, option, default):
        value = self._cp.get(section, option)
        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            log.error('"{0}" for option [{1}] {2} is not a valid list,'
                ' using default value of "{3}"'.format(value, section, option, default))
            return default

    def getDict(self, section, option, default):
        value = self._cp.get(section, option)
        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            log.error('"{0}" for option [{1}] {2} is not a valid dict,'
                ' using default value of "{3}"'.format(value, section, option, default))
            return default

    def __enter__(self):
        self.read()
        return self

    def __exit__(self, type, value, traceback):
        if self.require_write:
            self.write()


if __name__ == "__main__":

    with RuntimeConfig() as rc:
        print rc.get('TEST', 'STR_VALUE', 'default')
        rc.set('TEST', 'VALUE', 10)

    with RuntimeConfig('~/.axis_preferences') as rc:
        print rc.get('DEFAULT', 'recentfiles', default=[])



#!/usr/bin/env python

"""QCompile - compile .ui and .qrc files to Python

    Recursively compiles all Qt .ui and .qrc files in the
    specified package(s).

Usage:
  qcompile <package> ...
  qcompile -h

Example::

  $ qcompile .
  $ qcompile package1 package2/subpackage
"""

import os
import sys
import fnmatch
import subprocess

pyrcc = 'pyrcc5'
pyuic = 'pyuic5'

ok = "\033[32mok\033[0m"
error = "\033[31mERROR - unable to call {}\033[0m"


def compile(packages=['.',]):
    for package in packages:
        package_path = os.path.abspath(package)
        if not os.path.isdir(package_path):
            raise ValueError('Package "{}" not found!'.format(package))

        # Compile Qt UI files

        files = [os.path.join(dirpath, f)
            for dirpath, dirnames, files in os.walk(package_path)
            for f in files if f.endswith('.ui')]

        if len(files) > 0:
            print("Compiling .ui files in package '{}':".format(package))

            for infile in files:
                outfile = infile.replace('.ui', '_ui.py')

                sys.stdout.write("  {} => {} ... "
                                 .format(os.path.basename(infile),
                                         os.path.basename(outfile))
                                 )
                sys.stdout.flush()

                try:
                    ret = subprocess.call([pyuic, '-o', outfile, infile])
                except OSError:
                    print(error.format(pyuic))
                    break
                if ret == 0:
                    print(ok)

        # Compile Qt resource files
        files = [os.path.join(dirpath, f)
                 for dirpath, dirnames, files in os.walk(package_path)
                 for f in files if f.endswith('.qrc')]

        if len(files) > 0:
            print("\nCompiling .qrc files in package '{}':".format(package))

            for infile in files:
                outfile = infile.replace('.qrc', '_rc.py')

                sys.stdout.write("  {} => {} ... " \
                                 .format(os.path.basename(infile),
                                         os.path.basename(outfile))
                                 )
                sys.stdout.flush()

                try:
                    ret = subprocess.call([pyrcc, '-o', outfile, infile])
                except OSError:
                    print(error.format(pyrcc))
                    break
                if ret == 0:
                    print(ok)


def main():
    packages = sys.argv[1:]
    if len(packages) == 0 or '-h' in packages:
        print(__doc__)
    else:
        compile(packages=sys.argv[1:])


if __name__ == '__main__':
    main()



"""External Widgets

This module loads custom widgets defined in external packages
so they are available in QtDesigner.
"""

from pkg_resources import iter_entry_points

from qtpyvcp.utilities.logger import getLogger

LOG = getLogger(__name__)


for entry_point in iter_entry_points(group='qtpyvcp.widgets'):

    try:
        group_name = entry_point.name
        mod = entry_point.load()

        for name in dir(mod):
            if name.startswith('_'):
                continue

            try:
                plugin_cls = getattr(mod, name)
                plugin_cls.group_name = group_name
                globals()[name] = plugin_cls
            except Exception as e:
                LOG.exception("Error while loading custom widget plugin '%s'.",
                              name, exc_info=e)

    except Exception as e:
        LOG.exception("Error while trying to load custom widget package '%s'.",
                      group_name, exc_info=e)



import os
import hal

INI_FILE = os.getenv('INI_FILE_NAME')
TCLPATH = os.getenv('LINUXCNC_TCL_DIR', '/usr/lib/tcltk/linuxcnc')

"""Tool Actions launch LinuxCNC tools"""

def halshow():
    """Launch HALShow utility to view HAL and a Watch Window

    * Components
    * Pins
    * Parameters
    * Signals
    * Functions
    * Threads

    ActionButton syntax::

        tool_actions.halshow

    """
    p = os.popen("tclsh {0}/bin/halshow.tcl &".format(TCLPATH))

def calibration():
    """Launch the HAL PID calibration utility.

    Test PID, Scale, Acceleration and Velocity settings if they are in the INI
    file.

    ActionButton syntax::

        tool_actions.calibration

    """
    p = os.popen("tclsh {0}/bin/emccalib.tcl -ini {1} > /dev/null &"
                 .format(TCLPATH, INI_FILE), "w")

def halmeter():
    """Launch the HALMeter utility to display the current value of a single pin.

    ActionButton syntax::

        tool_actions.halmeter

    """
    p = os.popen("halmeter &")

def status():
    """Launch the LinuxCNC status monitor utility.

    ActionButton syntax::

        tool_actions.status

    """
    p = os.popen("linuxcnctop  > /dev/null &", "w")

def halscope():
    """Launch the HALScope utility.

    Halscope is an oscilloscope for the HAL

    ActionButton syntax::

        tool_actions.halscope

    """
    p = os.popen("halscope  > /dev/null &", "w")

def classicladder():
    """Launch the ClassicLadder editor.

    Todo:
        classicladder.bindOk should check for classicladder comp
    """
    if hal.component_exists("classicladder_rt"):
        p = os.popen("classicladder  &", "w")
    else:
        text = "Classicladder real-time component not detected"
        print text
        # self.error_dialog.run(text)

def simulate_probe():
    """Launch Simulate Probe

    ActionButton syntax::

        tool_actions.simulate_probe

    """
    p = os.popen("simulate_probe > /dev/null &", "w")

halshow.ok = calibration.ok = halmeter.ok = status.ok = halscope.ok = classicladder.ok = simulate_probe.ok = lambda widget: True
halshow.bindOk = calibration.bindOk = halmeter.bindOk = status.bindOk = halscope.bindOk = classicladder.bindOk = simulate_probe.bindOk = lambda widget: None



"""
HAL Interface
-------------

This module allows the creation of userspace HAL components in Python.
This includes pins and parameters of the various HAL types.

Example:

.. code-block:: python

    from qtpyvcp import hal

    # create a new component and add some pins
    comp = hal.getComponent("loop-back")
    comp.addPin("in", "float", "in")
    comp.addPin("out", "float", "out")

    # mark the component as 'ready'
    comp.ready()

    # define a function to call when the input pin changes
    def onInChanged(new_value):
        # loop the out pin to the in pin value
        comp.getPin('out').value = new_value

    # connect the listener to the input pin
    comp.addListener('in', onInChanged)

"""

from qtpyvcp.utilities.logger import getLogger

from hal_qlib import QComponent, QPin

COMPONENTS = {}
LOG = getLogger(__name__)


def component(name):
    """Initializes a new HAL component and registers it."""
    comp = QComponent(name)
    COMPONENTS[name] = comp
    return comp


def getComponent(name='qtpyvcp'):
    """Get HAL component.

    Args:
        name (str) : The name of the component to get. Defaults to `qtpyvcp`.

    Returns:
        QComponent : An existing or new HAL component.
    """

    try:
        comp = COMPONENTS[name]
        LOG.debug("Using existing HAL component: %s", name)
    except KeyError:
        LOG.info("Creating new HAL component: %s", name)
        comp = component(name)

    return comp


if __name__ == "__main__":

    from qtpy.QtWidgets import QApplication
    from qtpyvcp import hal

    app = QApplication([])

    # create a new component and add some pins
    comp = hal.getComponent("loop-back")
    comp.addPin("in", "float", "in")
    comp.addPin("out", "float", "out")

    # mark the component as 'ready'
    comp.ready()

    # define a function to call when the input pin changes
    def onInChanged(new_value):
        print("loop-back.in pin changed:", new_value)
        # loop the out pin to the in pin value
        comp.getPin('out').value = new_value

    # connect the listener to the input pin
    comp.addListener('in', onInChanged)

    app.exec_()



# Copyright (c) 2017-2018, SLAC National Accelerator Laboratory

# This file has been adapted from PyDM, and can be redistributed and/or
# modified in accordance with terms in conditions set forth in the BSD
# 3-Clause License. You can find the complete licence text in the LICENCES
# directory.

# Links:
#   PyDM Project: https://github.com/slaclab/pydm
#   PyDM Licence: https://github.com/slaclab/pydm/blob/master/LICENSE.md

import os
import json
import functools
import webbrowser

from qtpy import uic
from qtpy import QtWidgets, QtCore, QtDesigner

from qtpyvcp import PLUGINS
from qtpyvcp.plugins import DataChannel, getPlugin
from qtpyvcp.utilities.settings import Setting
from plugin_extension import _PluginExtension

# Set up logging
from qtpyvcp.utilities import logger
LOG = logger.getLogger(__name__)

RULE_PROPERTIES = {
    'None': ['None', None],
    'Enable': ['setEnabled', bool],
    'Visible': ['setVisible', bool],
    'Style Class': ['setStyleClass', str],
    # 'Opacity': ['setOpacity', float]
}


class RulesEditorExtension(_PluginExtension):
    def __init__(self, widget):
        super(RulesEditorExtension, self).__init__(widget)
        self.widget = widget
        self.addTaskMenuAction("Edit Widget Rules...", self.editAction)

    def editAction(self, state):
        RulesEditor(self.widget, parent=None).exec_()


class ChanInfoDialog(QtWidgets.QDialog):
    def __init__(self, info, parent=None):
        super(ChanInfoDialog, self).__init__(parent)
        ui_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               "channel_info_dialog.ui")
        uic.loadUi(ui_file, self)

        ch_obj, ch_exp, ch_val, ch_doc = info

        doc_lines = [l.strip() for l in ch_doc.strip().splitlines()]
        title = doc_lines[0]

        self.lbl_rtn_typ.setText(type(ch_obj.getValue()).__name__)

        for line in doc_lines:
            if line.startswith(':returns:'):
                self.lbl_chan_rtn.setText(line.replace(':returns:', '').strip())
            elif line.startswith(':rtype:'):
                self.lbl_rtn_typ.setText(line.replace(':rtype:', '').strip())

        self.lbl_chan_name.setText(title)
        self.tb_chan_doc.setText(ch_doc)
        val = ch_obj.getValue()
        txt = ''
        if len(self.lbl_rtn_typ.text().split(',')) > 1:
            txt = ', ' + str(ch_val)
        self.lbl_can_val.setText(str(val) + txt)


class TableCheckButton(QtWidgets.QWidget):
    def __init__(self, checked=False):
        super(TableCheckButton, self).__init__()
        self.chk_bx = QtWidgets.QCheckBox()
        self.chk_bx.setChecked(checked)
        lay_out = QtWidgets.QHBoxLayout(self)
        lay_out.addWidget(self.chk_bx)
        lay_out.setAlignment(QtCore.Qt.AlignCenter)
        lay_out.setContentsMargins(0, 0, 0, 0)
        self.setLayout(lay_out)

    def __getattr__(self, attr):
        return getattr(self.chk_bx, attr)


class CompleterDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent=None):
        super(CompleterDelegate, self).__init__(parent)

        items = []
        for plugin, obj in PLUGINS.iteritems():
            for chan_name in obj.channels:
                items.append('{}:{}'.format(plugin, chan_name))

        self.completer = QtWidgets.QCompleter(sorted(items))
        self.completer.setCompletionColumn(0)
        self.completer.setCompletionRole(QtCore.Qt.EditRole)
        self.completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)

    def createEditor(self, parent, option, index):
        editor = QtWidgets.QLineEdit(parent)
        editor.setFrame(False)
        editor.setCompleter(self.completer)
        return editor


class RulesEditor(QtWidgets.QDialog):
    """QDialog for user-friendly editing of widget Rules in Qt Designer.

    Args:
        widget (QtPyVCPWidget) : The widget for which we want to edit the
            `rules` property.
    """

    def __init__(self, widget, parent=None):
        super(RulesEditor, self).__init__(parent)

        self.widget = widget
        self.app = QtWidgets.QApplication.instance()

        self.lst_rule_item = None
        self.loading_data = False

        self.available_properties = widget.RULE_PROPERTIES
        self.default_property = widget.DEFAULT_RULE_PROPERTY

        self.setup_ui()

        try:
            self.rules = json.loads(widget.rules)
        except:
            self.rules = []

        for ac in self.rules:
            self.lst_rules.addItem(ac.get("name", ''))

    def setup_ui(self):
        """Create the UI elements for the form."""

        self.setWindowTitle("QtPyVCP Widget Rules Editor")
        vlayout = QtWidgets.QVBoxLayout()
        vlayout.setContentsMargins(5, 5, 5, 5)
        vlayout.setSpacing(5)
        self.setLayout(vlayout)

        hlayout = QtWidgets.QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setSpacing(5)
        vlayout.addLayout(hlayout)

        # Creating the widgets for the String List and
        # buttons to add and remove actions
        list_frame = QtWidgets.QFrame(parent=self)
        list_frame.setMinimumHeight(300)
        list_frame.setMinimumWidth(240)
        list_frame.setLineWidth(1)
        list_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        list_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        lf_layout = QtWidgets.QVBoxLayout()
        list_frame.setLayout(lf_layout)

        lf_btn_layout = QtWidgets.QHBoxLayout()
        lf_btn_layout.setContentsMargins(0, 0, 0, 0)
        lf_btn_layout.setSpacing(5)

        self.btn_add_rule = QtWidgets.QPushButton(parent=self)
        self.btn_add_rule.setAutoDefault(False)
        self.btn_add_rule.setDefault(False)
        self.btn_add_rule.setText("Add Rule")
        self.btn_add_rule.clicked.connect(self.add_rule)

        self.btn_del_rule = QtWidgets.QPushButton(parent=self)
        self.btn_del_rule.setAutoDefault(False)
        self.btn_del_rule.setDefault(False)
        self.btn_del_rule.setText("Remove Rule")
        self.btn_del_rule.clicked.connect(self.del_rule)

        lf_btn_layout.addWidget(self.btn_add_rule)
        lf_btn_layout.addWidget(self.btn_del_rule)

        lf_layout.addLayout(lf_btn_layout)

        self.lst_rules = QtWidgets.QListWidget()
        self.lst_rules.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding))
        self.lst_rules.itemSelectionChanged.connect(self.load_from_list)
        lf_layout.addWidget(self.lst_rules)

        hlayout.addWidget(list_frame)

        # open help button
        btn_help = QtWidgets.QPushButton()
        btn_help.setAutoDefault(False)
        btn_help.setDefault(False)
        btn_help.setText("Help")
        btn_help.setStyleSheet("background-color: rgb(176, 227, 255);")
        btn_help.clicked.connect(functools.partial(self.open_help, open=True))

        # save button
        save_btn = QtWidgets.QPushButton("Save", parent=self)
        save_btn.setAutoDefault(False)
        save_btn.setDefault(False)
        save_btn.clicked.connect(self.saveChanges)

        # cancel button
        cancel_btn = QtWidgets.QPushButton("Cancel", parent=self)
        cancel_btn.setAutoDefault(False)
        cancel_btn.setDefault(False)
        cancel_btn.clicked.connect(self.cancelChanges)

        # add buttons to layout
        buttons_layout = QtWidgets.QHBoxLayout()
        buttons_layout.addWidget(btn_help)
        buttons_layout.addStretch()
        buttons_layout.addWidget(cancel_btn)
        buttons_layout.addWidget(save_btn)

        vlayout.addLayout(buttons_layout)

        # Creating the widgets that we will use to compose the
        # rule parameters
        self.frm_edit = QtWidgets.QFrame()
        self.frm_edit.setEnabled(False)
        self.frm_edit.setLineWidth(1)
        self.frm_edit.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frm_edit.setFrameShape(QtWidgets.QFrame.StyledPanel)

        frm_edit_layout = QtWidgets.QVBoxLayout()
        self.frm_edit.setLayout(frm_edit_layout)

        hlayout.addWidget(self.frm_edit)

        edit_name_layout = QtWidgets.QFormLayout()
        edit_name_layout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        lbl_name = QtWidgets.QLabel("Rule Name:")
        self.txt_name = QtWidgets.QLineEdit()
        self.txt_name.editingFinished.connect(self.name_changed)
        edit_name_layout.addRow(lbl_name, self.txt_name)
        lbl_property = QtWidgets.QLabel("Property:")
        self.cmb_property = QtWidgets.QComboBox()
        for name, prop in self.available_properties.items():
            self.cmb_property.addItem(name, prop)
        self.cmb_property.model().sort(0)
        edit_name_layout.addRow(lbl_property, self.cmb_property)

        frm_edit_layout.addLayout(edit_name_layout)

        btn_add_remove_layout = QtWidgets.QHBoxLayout()
        self.btn_add_channel = QtWidgets.QPushButton()
        self.btn_add_channel.setAutoDefault(False)
        self.btn_add_channel.setDefault(False)
        self.btn_add_channel.setText("Add Channel")
        self.btn_add_channel.setIconSize(QtCore.QSize(16, 16))
        self.btn_add_channel.clicked.connect(self.add_channel)
        self.btn_del_channel = QtWidgets.QPushButton()
        self.btn_del_channel.setAutoDefault(False)
        self.btn_del_channel.setDefault(False)
        self.btn_del_channel.setText("Remove Channel")
        self.btn_del_channel.setIconSize(QtCore.QSize(16, 16))
        self.btn_del_channel.clicked.connect(self.del_channel)
        btn_add_remove_layout.addWidget(self.btn_add_channel)
        btn_add_remove_layout.addWidget(self.btn_del_channel)

        frm_edit_layout.addLayout(btn_add_remove_layout)

        self.tbl_channels = QtWidgets.QTableWidget()
        self.tbl_channels.setMinimumWidth(350)
        self.tbl_channels.setShowGrid(True)
        self.tbl_channels.setCornerButtonEnabled(False)
        delegate = CompleterDelegate(self.tbl_channels)
        self.tbl_channels.setItemDelegateForColumn(0, delegate)
        self.tbl_channels.model().dataChanged.connect(self.tbl_channels_changed)
        headers = ["Channel", "Trigger", "Type"]
        self.tbl_channels.setColumnCount(len(headers))
        self.tbl_channels.setHorizontalHeaderLabels(headers)
        header = self.tbl_channels.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeToContents)
        # header.setSectionResizeMode(3, QtWidgets.QHeaderView.ResizeToContents)
        self.tbl_channels.setColumnWidth(3, 8)

        frm_edit_layout.addWidget(self.tbl_channels)

        expression_layout = QtWidgets.QFormLayout()
        expression_layout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        lbl_expected = QtWidgets.QLabel("Exp Type:")
        self.lbl_expected_type = QtWidgets.QLabel(parent=self)
        # self.lbl_expected_type.setText("")
        self.lbl_expected_type.setStyleSheet("color: rgb(0, 128, 255); font-weight: bold;")

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.lbl_expected_type)
        hbox.addStretch()
        self.btn_chan_info = QtWidgets.QPushButton("Info ...")
        self.btn_chan_info.setAutoDefault(False)
        self.btn_chan_info.pressed.connect(self.show_chan_info)
        hbox.addWidget(self.btn_chan_info)

        expression_layout.addRow(lbl_expected, hbox)

        lbl_expression = QtWidgets.QLabel("Expression:")
        expr_help_layout = QtWidgets.QHBoxLayout()
        self.txt_expression = QtWidgets.QLineEdit()
        self.txt_expression.editingFinished.connect(self.expression_changed)
        expr_help_layout.addWidget(self.txt_expression)
        expression_layout.addRow(lbl_expression, expr_help_layout)

        self.cmb_property.currentIndexChanged.connect(self.property_changed)
        self.cmb_property.setCurrentText(self.default_property)

        frm_edit_layout.addLayout(expression_layout)

    def clear_form(self):
        """Clear the form and reset the fields."""
        self.loading_data = True
        self.lst_rule_item = None
        self.txt_name.setText("")
        self.cmb_property.setCurrentIndex(-1)
        self.tbl_channels.clearContents()
        self.txt_expression.setText("")
        self.frm_edit.setEnabled(False)
        self.loading_data = False

    def load_from_list(self):
        """Load an entry from the list into the editing form."""
        item = self.lst_rules.currentItem()
        idx = self.lst_rules.indexFromItem(item).row()

        if idx < 0:
            return

        self.loading_data = True
        self.lst_rule_item = item
        data = self.rules[idx]
        self.txt_name.setText(data.get('name', ''))
        prop = data.get('property', '')
        self.cmb_property.setCurrentText(prop)
        self.property_changed(0)
        self.txt_expression.setText(data.get('expression', ''))
        self.txt_expression.setEnabled(
            self.available_properties.get(prop, ['None', None])[1] is not None)


        channels = data.get('channels', [])
        self.tbl_channels.clearContents()
        self.tbl_channels.setRowCount(len(channels))
        vlabel = [str(i) for i in range(len(channels))]
        self.tbl_channels.setVerticalHeaderLabels(vlabel)
        for row, ch in enumerate(channels):
            ch_name = ch.get('url', '')
            ch_tr = ch.get('trigger', False)
            ch_obj, ch_exp, ch_val, ch_desc = self.get_channel_data(ch_name)
            self.tbl_channels.setItem(row, 0,
                                      QtWidgets.QTableWidgetItem(str(ch_name)))

            tr_chk = TableCheckButton(checked=ch_tr)
            self.tbl_channels.setCellWidget(row, 1, tr_chk)

            typ_lbl = QtWidgets.QLabel()
            typ_lbl.setAlignment(QtCore.Qt.AlignCenter)

            if ch_val is None:
                typ_lbl.setText("<font color='red'>error</font>")
            else:
                typ_lbl.setText("<font color='green'>{}</font>".format(type(ch_val).__name__))
            self.tbl_channels.setCellWidget(row, 2, typ_lbl)

        self.frm_edit.setEnabled(True)
        self.loading_data = False

    def add_rule(self):
        """Add a new rule to the list of rules."""
        default_name = "New Rule"
        data = {"name": default_name,
                "property": self.default_property,
                "expression": "",
                "channels": []
                }
        self.rules.append(data)
        self.lst_rule_item = QtWidgets.QListWidgetItem()
        self.lst_rule_item.setText(default_name)
        self.lst_rules.addItem(self.lst_rule_item)
        self.lst_rules.setCurrentItem(self.lst_rule_item)
        self.load_from_list()
        self.txt_name.setFocus()

    def get_current_index(self):
        """
        Calculate and return the selected index from the list of rules.

        Returns:
            int : The index selected in the rules list or -1 if the item
                does not exist.
        """
        if self.lst_rule_item is None:
            return -1
        return self.lst_rules.indexFromItem(self.lst_rule_item).row()

    def change_entry(self, entry, value):
        """Change an entry in the rules dictionary.

        Args:
            entry (str) : The key for the dictionary
            value (any) : The value to set on the key. It can be any type,
                depending on the key.
        """
        idx = self.get_current_index()
        self.rules[idx][entry] = value

    def del_rule(self):
        """Delete the rule selected in the rules list."""
        idx = self.get_current_index()
        if idx < 0:
            return
        name = self.lst_rule_item.text()

        confirm_message = "Are you sure you want to delete Rule: {}?".format(
            name)
        reply = QtWidgets.QMessageBox().question(self, 'Message',
                                             confirm_message,
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            self.lst_rules.takeItem(idx)
            self.lst_rules.clearSelection()
            self.rules.pop(idx)
            self.clear_form()

    def add_channel(self):
        """Add a new empty channel to the table."""
        self.loading_data = True

        # Make the first entry be checked as suggestion
        if self.tbl_channels.rowCount() == 0:
            state = QtCore.Qt.Checked
        else:
            state = QtCore.Qt.Unchecked

        self.tbl_channels.insertRow(self.tbl_channels.rowCount())
        row = self.tbl_channels.rowCount() - 1
        self.tbl_channels.setItem(row, 0, QtWidgets.QTableWidgetItem(""))
        checkBoxItem = TableCheckButton(checked=state)
        self.tbl_channels.setCellWidget(row, 1, checkBoxItem)
        typ_lbl = QtWidgets.QLabel()
        typ_lbl.setAlignment(QtCore.Qt.AlignCenter)
        self.tbl_channels.setCellWidget(row, 2, typ_lbl)
        vlabel = [str(i) for i in range(self.tbl_channels.rowCount())]
        self.tbl_channels.setVerticalHeaderLabels(vlabel)
        self.loading_data = False
        self.update_channels()

    def del_channel(self):
        """Delete the selected channel at the table."""
        items = self.tbl_channels.selectedIndexes()
        if len(items) == 0:
            return

        c = "channel" if len(items) == 1 else "channels"
        confirm_message = "Delete the selected {}?".format(c)
        reply = QtWidgets.QMessageBox().question(self, 'Message',
                                             confirm_message,
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            for itm in reversed(items):
                row = itm.row()
                self.tbl_channels.removeRow(row)

        self.update_channels()

    def open_help(self, open=True):
        """Open the Help context for Rules.

        The documentation website prefix is given by the `QTPYVCP_DOCS_URL`
        environmnet variable. If not defined it defaults to
        `https://kcjengr.github.io/qtpyvcp`

        Args:
            open (bool) : Whether or not to open the page in a web browser.
        """
        docs_url = os.getenv("QTPYVCP_DOCS_URL", None)
        if docs_url is None:
            docs_url = "https://kcjengr.github.io/qtpyvcp"
        expression_url = "tutorials/widget_rules.html"
        help_url = "{}/{}".format(docs_url, expression_url)
        if open:
            webbrowser.open(help_url, new=2, autoraise=True)
        else:
            return help_url

    def show_chan_info(self):
        selection = self.tbl_channels.selectedIndexes()
        if len(selection) > 0:
            row = selection[0].row()
            model = self.tbl_channels.model()
            chan = model.data(model.index(row, 0))
            info = self.get_channel_data(chan)
            ChanInfoDialog(info).exec_()

    def name_changed(self):
        """Callback executed when the rule name is changed."""
        self.lst_rule_item.setText(self.txt_name.text())
        self.change_entry("name", self.txt_name.text())

    def property_changed(self, index):
        """Callback executed when the property is selected."""
        try:
            prop = self.cmb_property.currentData()
            if prop[1] is None:
                self.lbl_expected_type.setText("None")
                self.txt_expression.setEnabled(False)
                self.txt_expression.setToolTip("Expression not used with the 'None' property rule.")
            else:
                self.lbl_expected_type.setText(prop[1].__name__)
                self.txt_expression.setEnabled(True)
            idx = self.get_current_index()
            self.change_entry("property", self.cmb_property.currentText())
        except Exception as e:
            print "error", e
            self.lbl_expected_type.setText("")

    def tbl_channels_changed(self, table_item):
        """Callback executed when the channels in the table are modified."""
        if self.loading_data:
            return

        row = table_item.row()
        ch_name = table_item.data()
        ch_obj, ch_exp, ch_val, ch_desc = self.get_channel_data(ch_name)

        typ_lbl = self.tbl_channels.cellWidget(row, 2)

        if isinstance(ch_obj, (DataChannel, Setting)) and ch_val is not None:
            typ_lbl.setText("<font color='green'>{}</font>"
                            .format(type(ch_val).__name__))
        else:
            typ_lbl.setText("<font color='red'>error</font>")

        self.update_channels()

    def update_channels(self):
        """Update the JSON format chanels to match the data in the table."""

        new_channels = []

        for row in range(self.tbl_channels.rowCount()):
            ch = self.tbl_channels.item(row, 0).text()
            tr = self.tbl_channels.cellWidget(row, 1).isChecked()
            new_channels.append({"url": ch, "trigger": tr})

        self.change_entry("channels", new_channels)

    def get_channel_data(self, url):

        protocol, sep, item = url.partition(':')
        try:
            plugin = getPlugin(protocol)
        except ValueError:
            return None, None, None, None

        chan_obj, chan_exp = plugin.getChannel(item)
        if chan_obj is not None:
            chan_val = chan_exp()
        else:
            chan_val = None

        return chan_obj, chan_exp, chan_val, chan_obj.__doc__

    def expression_changed(self):
        """Callback executed when the expression is modified."""
        self.change_entry("expression", self.txt_expression.text())

    def is_data_valid(self):
        """
        Sanity check the form data.

        Returns:
            tuple (bool, str) : True and "" in case there is no error or False
                and the error message otherwise.
        """
        errors = []
        for idx, rule in enumerate(self.rules):
            name = rule.get("name")
            prop = rule.get("property")
            expression = rule.get("expression")
            channels = rule.get("channels", [])

            channel_values = []

            if name is None or name == "":
                errors.append("Rule #{} has no name.".format(idx))

            if len(channels) == 0:
                errors.append("Rule #{} has no channel.".format(idx))

            if self.available_properties[prop][1] is None:
                # No need to check anything else.
                break

            if expression is None or expression == "":
                errors.append("Rule #{} has no expression.".format(idx))
            else:
                found_trigger = False
                for ch_idx, ch in enumerate(channels):

                    if not ch.get("url", ""):
                        errors.append("Rule #{} - Ch. #{} has no channel.".format(idx, ch_idx))

                    if ch.get("trigger", False) and not found_trigger:
                        found_trigger = True

                    # get chan values for use when checking expression
                    ch_obj, ch_exp, ch_val, ch_desc = self.get_channel_data(ch.get("url", ""))

                    if isinstance(ch_obj, (DataChannel, Setting)):
                        channel_values.append(ch_exp())
                    else:
                        errors.append("Rule #{} is not a valid channel.".format(idx))
                        continue

                if not found_trigger:
                    errors.append(
                        "Rule #{} has no trigger channel.".format(idx))

            try:
                # check python expression
                value = eval(expression, {'ch': channel_values})

                # check return type
                exp_typ = self.available_properties[prop][1]
                act_typ = type(value)
                if act_typ != exp_typ:
                    errors.append("Rule #{} expression returned '{}', but expected '{}'."
                                  .format(idx + 1, act_typ.__name__, exp_typ.__name__))

            except:
                LOG.exception("Error evaluating Rule #{} expression.".format(idx))
                errors.append("Rule #{} expression is not valid.".format(idx))

        if len(errors) > 0:
            error_msg = os.linesep.join(errors)
            return False, error_msg

        return True, ""

    @QtCore.Slot()
    def saveChanges(self):
        """Save the new rules at the widget `rules` property."""
        # If the form is being edited, we make sure self.rules has all the
        # latest values from the form before we try to validate.  This fixes
        # a problem where the last form item changed wouldn't get saved unless
        # the user knew to hit 'enter' or leave the field to end editing before
        # hitting save.
        if self.frm_edit.isEnabled():
            self.expression_changed()
            self.name_changed()
            self.update_channels()
        is_valid, message = self.is_data_valid()
        if is_valid:
            data = json.dumps(self.rules)
            print json.dumps(self.rules, sort_keys=True, indent=4)
            formWindow = QtDesigner.QDesignerFormWindowInterface.findFormWindow(self.widget)
            if formWindow:
                formWindow.cursor().setProperty("rules", data)
            self.accept()
        else:
            QtWidgets.QMessageBox.critical(self, "Error Saving", message,
                                       QtWidgets.QMessageBox.Ok)

    @QtCore.Slot()
    def cancelChanges(self):
        """Abort the changes and close the dialog."""
        self.close()




import os
from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QApplication

from qtpy.QtDesigner import QDesignerFormWindowInterface


from qtpyvcp.widgets.qtdesigner import _PluginExtension

UI_FILE = os.path.join(os.path.dirname(__file__), "dro_editor.ui")


class DroEditorExtension(_PluginExtension):
    def __init__(self, widget):
        super(DroEditorExtension, self).__init__(widget)
        self.widget = widget
        self.addTaskMenuAction("Edit DRO Settings ...", self.editAction)

    def editAction(self, state):
        DroEditor(self.widget, parent=None).exec_()


class DroEditor(QDialog):
    """QDialog for user-friendly editing of DRO properties in Qt Designer."""

    def __init__(self, widget, parent=None):
        super(DroEditor, self).__init__(parent)

        self.widget = widget
        self.app = QApplication.instance()

        uic.loadUi(UI_FILE, self)

        self.axisCombo.setCurrentIndex(self.widget.axisNumber)
        self.refTypCombo.setCurrentIndex(self.widget.referenceType)

        self.inFmtEntry.setText(self.widget.inchFormat)
        self.mmFmtEntry.setText(self.widget.millimeterFormat)
        self.degFmtEntry.setText(self.widget.degreeFormat)

        self.latheModeCombo.setCurrentIndex(self.widget.latheMode)

        bb = self.buttonBox
        bb.button(QDialogButtonBox.Apply).setDefault(True)
        bb.button(QDialogButtonBox.Cancel).setDefault(False)
        bb.button(QDialogButtonBox.Apply).clicked.connect(self.accept)

    @Slot()
    def accept(self):
        """Commit changes"""
        # general options
        self.setCursorProperty('axisNumber', self.axisCombo.currentIndex())
        self.setCursorProperty('referenceType', self.refTypCombo.currentIndex())

        # format options
        self.setCursorProperty('inchFormat', self.inFmtEntry.text())
        self.setCursorProperty('millimeterFormat', self.mmFmtEntry.text())
        self.setCursorProperty('degreeFormat', self.degFmtEntry.text())

        # lathe options
        self.setCursorProperty('latheMode', self.latheModeCombo.currentIndex())

        self.close()

    def setCursorProperty(self, prop_name, value):
        form = QDesignerFormWindowInterface.findFormWindow(self.widget)
        if form:
            form.cursor().setProperty(prop_name, value)



#   Copyright (c) 2018 Kurt Jacobson
#      <kurtcjacobson@gmail.com>
#
#   This file is part of QtPyVCP.
#
#   QtPyVCP is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 2 of the License, or
#   (at your option) any later version.
#
#   QtPyVCP is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with QtPyVCP.  If not, see <http://www.gnu.org/licenses/>.

import os
import linuxcnc
INIFILE = linuxcnc.ini(os.getenv("INI_FILE_NAME"))

from qtpy.QtWidgets import QLabel
from qtpy.QtCore import Slot, Property

from qtpyvcp.plugins import getPlugin

STATUS = getPlugin('status')
POSITION = getPlugin('position')

from qtpyvcp.widgets import VCPWidget

from qtpyvcp.utilities import logger

LOG = logger.getLogger(__name__)

IN_DESIGNER = os.getenv('DESIGNER') != None


from qtpyvcp.lib.decorators import deprecated

class Axis(object):
    ALL = -1
    X, Y, Z, A, B, C, U, V, W = range(9)


class Units(object):
    Program = 0  # Use program units
    Inch = 1  # CANON_UNITS_INCHES=1
    Metric = 2  # CANON_UNITS_MM=2


class RefType(object):
    Absolute = 0
    Relative = 1
    DistanceToGo = 2

    @classmethod
    def toString(self, ref_type):
        return ['abs', 'rel', 'dtg'][ref_type]


@deprecated(reason='new DRO implementation', replaced_by='DroLabel')
class DROWidget(QLabel, VCPWidget, Axis, RefType, Units):
    from PyQt5.QtCore import Q_ENUMS
    Q_ENUMS(Axis)
    Q_ENUMS(RefType)
    Q_ENUMS(Units)

    def __init__(self, parent=None):
        super(DROWidget, self).__init__(parent)

        self.log = logger.getLogger(__name__)

        self._axis_number = Axis.X
        self._ref_typ = RefType.Absolute

        self._metric_format = '%10.3f'
        self._imperial_format = '%9.4f'

        units = INIFILE.find("TRAJ", "LINEAR_UNITS") or "inch"
        if units == 'inch':
            self._format = self._imperial_format
        else:
            self._format = self._metric_format

        self.update(getattr(POSITION, RefType.toString(self._ref_typ)).getValue())
        STATUS.program_units.notify(self.onUnitsChanged, 'string')

    def update(self, pos):
        self.setText(self._format % pos[self._axis_number])

    def onUnitsChanged(self, units):
        if units == 'in':
            self._format = self._imperial_format
        else:
            self._format = self._metric_format
        self.update(getattr(POSITION, RefType.toString(self._ref_typ)).getValue())

    def initialize(self):
        getattr(POSITION, RefType.toString(self._ref_typ)).notify(self.update)

    # ==========================================================================
    # Designer property Getters/Setters
    # ==========================================================================

    @Property(RefType)
    def referenceType(self):
        return self._ref_typ

    @referenceType.setter
    def referenceType(self, ref_typ):
        new_ref_typ = RefType.toString(ref_typ)
        self._ref_typ = ref_typ
        self.update(getattr(POSITION, new_ref_typ).getValue())

    @Property(Axis)
    def axis(self):
        return self._axis_number

    @axis.setter
    def axis(self, axis):
        self._axis_number = axis
        self.update(getattr(POSITION, RefType.toString(self._ref_typ)).getValue())

    def getDiamterMode(self):
        return self._diameter_mode

    @Slot(bool)
    def setDiamterMode(self, diameter_mode):
        self._diameter_mode = diameter_mode
        self.updateUnits(STATUS.stat.program_units)

    @Property(str)
    def metricTemplate(self):
        return self._metric_format

    @metricTemplate.setter
    def metricTemplate(self, value):
        try:
            value % 12.456789
        except ValueError as va:
            self.log.debug(va)
            return
        self._metric_format = value
        self._format = self._metric_format
        self.update(getattr(POSITION, RefType.toString(self._ref_typ)).getValue())

    @Property(str)
    def imperialTemplate(self):
        return self._imperial_format

    @imperialTemplate.setter
    def imperialTemplate(self, value):
        try:
            value % 12.456789
        except ValueError as va:
            self.log.debug(va)
            return
        self._imperial_format = value
        self._format = self._imperial_format
        self.update(getattr(POSITION, RefType.toString(self._ref_typ)).getValue())


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    w = DROWidget()
    w.show()
    sys.exit(app.exec_())



import os

# Workarround for nvidia propietary drivers

import ctypes
import ctypes.util

ctypes.CDLL(ctypes.util.find_library("GL"), mode=ctypes.RTLD_GLOBAL)

# end of Workarround


from qtpy.QtCore import Signal, Slot, QUrl, QTimer
from qtpy.QtQuickWidgets import QQuickWidget

from qtpyvcp.plugins import getPlugin
from qtpyvcp.utilities import logger

LOG = logger.getLogger(__name__)
STATUS = getPlugin('status')
TOOLTABLE = getPlugin('tooltable')
IN_DESIGNER = os.getenv('DESIGNER', False)
WIDGET_PATH = os.path.dirname(os.path.abspath(__file__))


class DynATC(QQuickWidget):

    rotateSig = Signal(int, int, arguments=['steps', 'direction'])

    showToolSig = Signal(float, float, arguments=['pocket', 'tool_num'])
    hideToolSig = Signal(float, arguments=['pocket'])

    homeMsgSig = Signal(str, arguments=["message"])

    def __init__(self, parent=None):
        super(DynATC, self).__init__(parent)

        if IN_DESIGNER:
            return

        self.atc_position = 0
        self.pocket = 1
        self.home = 0
        self.homing = 0
        self.pocket_slots = 12

        self.engine().rootContext().setContextProperty("atc_spiner", self)
        qml_path = os.path.join(WIDGET_PATH, "atc.qml")
        url = QUrl.fromLocalFile(qml_path)

        self.setSource(url)  # Fixme fails on qtdesigner

        self.tool_table = None
        self.status_tool_table = None
        self.pockets = dict()
        self.tools = None

        for pocket in range(1, 13):
            self.hideToolSig.emit(pocket)

    def hideEvent(self, *args, **kwargs):
        pass  # hack to prevent animation glitch when we are on another tab FIXME

    def load_tools(self):
        # print("load_tools")
        for i in range(1, self.pocket_slots+1):
            self.hideToolSig.emit(i)

        for pocket, tool in self.pockets.items():
            if tool != 0:
                self.showToolSig.emit(pocket, tool)
            else:
                self.hideToolSig.emit(pocket)

    def store_tool(self, pocket, tool_num):
        self.pockets[pocket] = tool_num
        #
        # print(type(pocket), pocket)
        # print(type(tool_num), tool_num)
        if tool_num != 0:
            # print("show tool {} at pocket {}".format(tool_num, pocket))
            self.showToolSig.emit(pocket, tool_num)
        else:
            # print("Hide tool at pocket {}".format(pocket))
            self.hideToolSig.emit(pocket)

    def atc_message(self, msg=""):
        self.homeMsgSig.emit(msg)

    def rotate(self, steps, direction):
        if direction == "cw":
            self.rotateSig.emit(steps, 1)
        elif direction == "ccw":
            self.rotateSig.emit(steps, -1)



import os
import subprocess

import linuxcnc
import psutil

from pyudev.pyqt5 import MonitorObserver
from pyudev import Context, Monitor, Devices

from qtpy.QtCore import Slot, Property, Signal, QFile, QFileInfo, QDir, QIODevice
from qtpy.QtWidgets import QFileSystemModel, QComboBox, QTableView, QMessageBox, \
    QApplication, QAbstractItemView, QInputDialog, QLineEdit

from qtpyvcp.plugins import getPlugin
from qtpyvcp.actions.program_actions import load as loadProgram
from qtpyvcp.utilities.info import Info
from qtpyvcp.utilities.logger import getLogger
from qtpyvcp.lib.decorators import deprecated


LOG = getLogger(__name__)

IN_DESIGNER = os.getenv('DESIGNER') != None


class TableType(object):
    Local = 0
    Remote = 1


class RemovableDeviceComboBox(QComboBox):
    """ComboBox for choosing from a list of removable devices."""
    usbPresent = Signal(bool)
    currentPathChanged = Signal(str)
    currentDeviceEjectable = Signal(bool)

    def __init__(self, parent=None):
        super(RemovableDeviceComboBox, self).__init__(parent)

        self.setSizeAdjustPolicy(QComboBox.AdjustToContents)

        self._file_locations = getPlugin('file_locations')
        self._file_locations.removable_devices.notify(self.onRemovableDevicesChanged)
        self._file_locations.new_device.notify(self.onNewDeviceAdded)

        self.info = Info()
        self._program_prefix = self.info.getProgramPrefix()

        self.currentTextChanged.connect(self.onCurrentTextChanged)

        # initialize device list
        self.onRemovableDevicesChanged(self._file_locations.removable_devices.value)

    def showEvent(self, event=None):
        self.setCurrentText(self._file_locations.default_location)
        data = self.currentData() or {}
        self.currentDeviceEjectable.emit(data.get('removable', False))

    def onCurrentTextChanged(self, text):
        data = self.currentData()
        if data:
            self.currentPathChanged.emit(data.get('path', '/'))
            self.currentDeviceEjectable.emit(data.get('removable', False))

    def onRemovableDevicesChanged(self, devices):

        self.blockSignals(True)

        self.clear()

        for label, path in self._file_locations.local_locations.items():
            self.addItem(label, {'path': os.path.expanduser(path)})

        self.insertSeparator(100)

        if devices:
            for devices_node, device_data in devices.items():
                self.addItem(device_data.get('label', 'Unknown'), device_data)

        self.blockSignals(False)

    def onNewDeviceAdded(self, device):
        if device:
            self.setCurrentText(device.get('label'))
        else:
            self.setCurrentText(self._file_locations.default_location)

    @Slot()
    def ejectDevice(self):
        data = self.currentData()
        if data:
            self._file_locations.ejectDevice(data.get('device'))

class FileSystemTable(QTableView, TableType):

    if IN_DESIGNER:
        from PyQt5.QtCore import Q_ENUMS
        Q_ENUMS(TableType)

    gcodeFileSelected = Signal(bool)
    filePreviewText = Signal(str)
    fileNamePreviewText = Signal(str)
    transferFileRequest = Signal(str)
    rootChanged = Signal(str)
    atDeviceRoot = Signal(bool)

    def __init__(self, parent=None):
        super(FileSystemTable, self).__init__(parent)

        self._table_type = TableType.Local
        self._hidden_columns = ''

        # This prevents doing unneeded initialization
        # when QtDesginer loads the plugin.
        if parent is None:
            return

        self.parent = parent
        self.path_data = dict()

        self.selected_row = None
        self.clipboard = QApplication.clipboard()

        self.model = QFileSystemModel()
        self.model.setReadOnly(True)
        self.model.setFilter(QDir.AllDirs | QDir.NoDotAndDotDot | QDir.AllEntries)

        self.setModel(self.model)

        self.verticalHeader().hide()
        self.horizontalHeader().setStretchLastSection(True)
        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.selection_model = self.selectionModel()
        self.selection_model.selectionChanged.connect(self.onSelectionChanged)

        # open selected item on double click or enter pressed
        self.activated.connect(self.openSelectedItem)

        self.info = Info()
        self.nc_file_editor = self.info.getEditor()
        self.nc_file_dir = self.info.getProgramPrefix()
        self.nc_file_exts = self.info.getProgramExtentions()

        self.setRootPath(self.nc_file_dir)

        self.model.rootPathChanged.connect(self.onRootPathChanged)

    def showEvent(self, event=None):
        root_path = self.model.rootPath()
        self.rootChanged.emit(root_path)
        self.atDeviceRoot.emit(os.path.ismount(root_path))

    def onRootPathChanged(self, path):
        self.atDeviceRoot.emit(os.path.ismount(path))

    def onSelectionChanged(self, selected, deselected):

        if len(selected) == 0:
            return

        index = selected.indexes()[0]
        path = self.model.filePath(index)

        if os.path.isfile(path):
            self.gcodeFileSelected.emit(True)
            with open(path, 'r') as fh:
                content = fh.read()
            self.filePreviewText.emit(content)
            self.fileNamePreviewText.emit(path)
        else:
            self.gcodeFileSelected.emit(False)
            self.filePreviewText.emit('')
            self.fileNamePreviewText.emit('')

    @Slot()
    def openSelectedItem(self, index=None):
        """If ngc file, opens in LinuxCNC, if dir displays dir."""
        if index is None:
            selection = self.getSelection()
            if selection is None:
                return
            index = selection[0]

        path = self.model.filePath(self.rootIndex())
        name = self.model.filePath(index)

        absolute_path = os.path.join(path, name)

        file_info = QFileInfo(absolute_path)
        if file_info.isDir():
            self.model.setRootPath(absolute_path)
            self.setRootIndex(self.model.index(absolute_path))
            self.rootChanged.emit(absolute_path)

        elif file_info.isFile():
            # if file_info.completeSuffix() not in self.nc_file_exts:
            #     LOG.warn("Unsuported NC program type with extention .%s",
            #              file_info.completeSuffix())
            loadProgram(absolute_path)

    @Slot()
    def editSelectedFile(self):
        """Open the selected file in editor."""
        selection = self.getSelection()
        if selection is not None:
            path = self.model.filePath(selection[0])
            subprocess.Popen([self.nc_file_editor, path])
        return False

    @Slot()
    def loadSelectedFile(self):
        """Loads the selected file into LinuxCNC."""
        selection = self.getSelection()
        if selection is not None:
            path = self.model.filePath(selection[0])
            loadProgram(path)
            return True
        return False

    @Slot()
    def selectPrevious(self):
        """Select the previous item in the view."""
        selection = self.getSelection()
        if selection is None:
            # select last item in view
            self.selectRow(self.model.rowCount(self.rootIndex()) - 1)
        else:
            self.selectRow(selection[0].row() - 1)
        return True

    @Slot()
    def selectNext(self):
        """Select the next item in the view."""
        selection = self.getSelection()
        if selection is None:
            # select first item in view
            self.selectRow(0)
        else:
            self.selectRow(selection[-1].row() + 1)
        return True

    @Slot()
    def rename(self):
        """renames the selected file or folder"""
        index = self.selectionModel().currentIndex()
        path = self.model.filePath(index)
        if path:
            file_info = QFileInfo(path)

            if file_info.isFile():
                filename = self.rename_dialog("file")

                if filename:
                    q_file = QFile(path)
                    file_info.absolutePath()
                    new_path = os.path.join(file_info.absolutePath(), str(filename))
                    q_file.rename(new_path)

            elif file_info.isDir():
                filename = self.rename_dialog("directory")

                if filename:
                    directory = QDir(path)
                    file_info.absolutePath()
                    new_path = os.path.join(file_info.absolutePath(), str(filename))
                    directory.rename(path, new_path)

    @Slot()
    def newFile(self):
        """Create a new empty file"""
        path = self.model.filePath(self.rootIndex())
        new_file_path = os.path.join(path, "New File.ngc")

        count = 1
        while os.path.exists(new_file_path):
            new_file_path = os.path.join(path, "New File {}.ngc".format(count))
            count += 1

        new_file = QFile(new_file_path)
        new_file.open(QIODevice.ReadWrite)

    @Slot()
    def newFolder(self):
        path = self.model.filePath(self.rootIndex())

        new_name = 'New Folder'

        count = 1
        while os.path.exists(os.path.join(path, new_name)):
            new_name = "New Folder {}".format(count)
            count += 1

        directory = QDir(path)
        directory.mkpath(new_name)
        directory.setPath(new_name)

    @Slot()
    @deprecated(replaced_by='newFolder',
                reason='for consistency with newFile method name')
    def createDirectory(self):
        self.newFolder()

    @Slot()
    def deleteItem(self):
        """Delete the selected item (either a file or folder)."""
        # ToDo: use Move2Trash, instead of deleting the file
        index = self.selectionModel().currentIndex()
        path = self.model.filePath(index)
        if path:
            file_info = QFileInfo(path)
            if file_info.isFile():
                if not self.ask_dialog("Do you wan't to delete the selected file?"):
                    return
                q_file = QFile(path)
                q_file.remove()

            elif file_info.isDir():
                if not self.ask_dialog("Do you wan't to delete the selected directory?"):
                    return
                directory = QDir(path)
                directory.removeRecursively()

    @Slot()
    @deprecated(replaced_by='deleteItem',
                reason='because of unclear method name')
    def deleteFile(self):
        self.deleteItem()

    @Slot(str)
    def setRootPath(self, root_path):
        """Sets the currently displayed path."""

        self.rootChanged.emit(root_path)
        self.model.setRootPath(root_path)
        self.setRootIndex(self.model.index(root_path))

        return True

    @Slot()
    def viewParentDirectory(self):
        """View the parent directory of the current view."""

        path = self.model.filePath(self.rootIndex())

        file_info = QFileInfo(path)
        directory = file_info.dir()
        new_path = directory.absolutePath()

        if os.path.ismount(path):
            return

        currentRoot = self.rootIndex()

        self.model.setRootPath(new_path)
        self.setRootIndex(currentRoot.parent())
        self.rootChanged.emit(new_path)


    @Slot()
    @deprecated(replaced_by='viewParentDirectory')
    def goUP(self):
        self.viewParentDirecotry()

    @Slot()
    def viewHomeDirectory(self):
        self.setRootPath(os.path.expanduser('~/'))

    @Slot()
    def viewNCFilesDirectory(self):
        # ToDo: Make preset user definable
        path = os.path.expanduser(self._nc_files_dir)
        self.setRootPath(path)

    @Slot()
    def viewPresetDirectory(self):
        # ToDo: Make preset user definable
        preset = os.path.expanduser(self._nc_files_dir)
        self.setRootPath(preset)

    @Slot()
    def doFileTransfer(self):
        index = self.selectionModel().currentIndex()
        path = self.model.filePath(index)
        self.transferFileRequest.emit(path)

    @Slot(str)
    def transferFile(self, src_path):
        dest_path = self.model.filePath(self.rootIndex())

        src_file = QFile()
        src_file.setFileName(src_path)

        src_file_info = QFileInfo(src_path)

        dst_path = os.path.join(dest_path, src_file_info.fileName())

        src_file.copy(dst_path)

    @Slot()
    def getSelection(self):
        """Returns list of selected indexes, or None."""
        selection = self.selection_model.selectedIndexes()
        if len(selection) == 0:
            return None
        return selection

    @Slot()
    def getCurrentDirectory(self):
        return self.model.rootPath()

    @Property(TableType)
    def tableType(self):
        return self._table_type

    @tableType.setter
    def tableType(self, table_type):
        self._table_type = table_type
        if table_type == TableType.Local:
            self.setRootPath(self.nc_file_dir)
        else:
            self.setRootPath('/media/')

    @Property(str)
    def hiddenColumns(self):
        """String of comma separated column numbers to hide."""
        return self._hidden_columns

    @hiddenColumns.setter
    def hiddenColumns(self, columns):
        try:
            col_list = [int(c) for c in columns.split(',') if c != '']
        except:
            return False

        self._hidden_columns = columns

        header = self.horizontalHeader()
        for col in range(4):
            if col in col_list:
                header.hideSection(col)
            else:
                header.showSection(col)

    def ask_dialog(self, message):
        box = QMessageBox.question(self.parent,
                                   'Are you sure?',
                                   message,
                                   QMessageBox.Yes,
                                   QMessageBox.No)
        if box == QMessageBox.Yes:
            return True
        else:
            return False

    def rename_dialog(self, data_type):
        text, ok_pressed = QInputDialog.getText(self.parent, "Rename", "New {} name:".format(data_type),
                                                QLineEdit.Normal, "")

        if ok_pressed and text != '':
            return text
        else:
            return False




from qtpy.QtWidgets import QLabel
from qtpy.QtCore import Property, Q_ENUMS

from qtpyvcp import hal
from qtpyvcp.widgets import HALWidget

from . import HalType


class HalLabel(QLabel, HALWidget, HalType):
    """HAL Label

    Label for displaying HAL pin values.

    Input pin type is selectable via the :class:`.pinType` property in designer,
    and can be any valid HAL type (bit, u32, s32, float).

    The text format can be specified via the :class:`.textFormat` property in
    designer and can be any valid python style format string.

    .. table:: Generated HAL Pins

        ========================= =========== =========
        HAL Pin Name              Type        Direction
        ========================= =========== =========
        qtpyvcp.label.enable      bit         in
        qtpyvcp.label.in          selecatable in
        ========================= =========== =========
    """
    Q_ENUMS(HalType)

    def __init__(self, parent=None):
        super(HalLabel, self).__init__(parent)

        self._in_pin = None
        self._enable_pin = None

        self._typ = "float"
        self._fmt = "%s"

        self.setValue(0)

    def setValue(self, value):
        self.setText(self._fmt % value)

    @Property(str)
    def textFormat(self):
        """Text Format Property

        Args:
            fmt (str) : A valid python style format string. Defaults to ``%s``.
        """
        return self._fmt

    @textFormat.setter
    def textFormat(self, fmt):
        self._fmt = fmt
        try:
            val = {'bit': False, 'u32': 0, 's32': 0, 'float': 0.0}[self._typ]
            self.setValue(val)
        except:
            pass

    @Property(HalType)
    def pinType(self):
        return getattr(HalType, self._typ)

    @pinType.setter
    def pinType(self, typ_enum):
        self._typ = HalType.toString(typ_enum)
        try:
            val = {'bit': False, 'u32': 0, 's32': 0, 'float': 0.0}[self._typ]
            self.setValue(val)
        except:
            pass

    def initialize(self):
        comp = hal.getComponent()
        obj_name = self.getPinBaseName()

        # add label.enable HAL pin
        self._enable_pin = comp.addPin(obj_name + ".enable", "bit", "in")
        self._enable_pin.value = self.isEnabled()
        self._enable_pin.valueChanged.connect(self.setEnabled)

        # add label.in HAL pin
        self._in_pin = comp.addPin(obj_name + ".in", self._typ, "in")
        self.setValue(self._in_pin.value)
        self._in_pin.valueChanged.connect(self.setValue)



import linuxcnc

# Set up logging
from qtpyvcp.utilities import logger
LOG = logger.getLogger(__name__)

from qtpyvcp.utilities.info import Info
from qtpyvcp.plugins import getPlugin

STATUS = getPlugin('status')
STAT = STATUS.stat
INFO = Info()

SPINDLES = range(INFO.spindles())
DEFAULT_SPEED = INFO.defaultSpindleSpeed()

CMD = linuxcnc.command()

from qtpyvcp.actions.base_actions import setTaskMode


def _spindle_exists(spindle):
    if spindle in SPINDLES:
        return True
    return False

def _spindle_ok(speed=None, spindle=0, widget=None):
    if spindle not in SPINDLES:
            ok = False
            msg = "No spindle No. {}".format(spindle)
    elif STAT.task_state != linuxcnc.STATE_ON:
        ok = False
        msg = "Power must be ON"
    elif STAT.task_mode == linuxcnc.MODE_AUTO:
        ok = False
        msg = "Mode must be MAN or MDI"
    else:
        ok = True
        msg = ""

    _spindle_ok.msg = msg

    if widget is not None:
        widget.setEnabled(ok)
        widget.setStatusTip(msg)
        widget.setToolTip(msg)

    return ok

def _spindle_bindOk(speed=None, spindle=0, widget=None):
    if not _spindle_exists(spindle):
        return
    STATUS.on.notify(lambda: _spindle_ok(spindle=spindle, widget=widget))
    STATUS.task_mode.onValueChanged(lambda: _spindle_ok(spindle=spindle, widget=widget))


def forward(speed=None, spindle=0):
    """Turn a spindle ON in the *FORWARD* direction.

    ActionButton syntax to start spindle 0 CW (spindle 0 is default)
    ::

        spindle.forward

    ActionButton syntax to start spindle 1 CW
    ::

        spindle.1.forward


    ActionButton syntax to start spindle 0 CW at 1800 RPM
    ::

        spindle.forward:1800

    Args:
        speed (int, optional) : The requested speed to spin the spindle at.
            If ``speed`` is not specified the current interpreter speed setting
            (as set by the last S word) is used, taking into account the
            value of the spindle override if it is enabled.
        spindle (int, optional) : The number of the spindle to turn ON. If
            ``spindle`` is not specified spindle 0 is assumed.
    """

    if speed is None:
        speed = getSpeed()
    CMD.spindle(linuxcnc.SPINDLE_FORWARD, speed, spindle)

def _spindle_forward_bindOk(speed=None, spindle=0, widget=None):
    if not _spindle_exists(spindle):
        return
    widget.setCheckable(True)
    STATUS.on.notify(lambda: _spindle_ok(spindle=spindle, widget=widget))
    STATUS.task_mode.onValueChanged(lambda: _spindle_ok(spindle=spindle, widget=widget))
    STATUS.spindle[spindle].direction.onValueChanged(lambda d: widget.setChecked(d == 1))

forward.ok = _spindle_ok
forward.bindOk = _spindle_forward_bindOk


def reverse(speed=None, spindle=0):
    """Turn a spindle ON in the *REVERSE* direction.

    ActionButton syntax to start spindle 0 CCW (spindle 0 is default)
    ::

        spindle.reverse

    ActionButton syntax to start spindle 1 CCW
    ::

        spindle.1.reverse


    ActionButton syntax to start spindle 0 CCW at 1800 RPM
    ::

        spindle.reverse:1800

    Args:
        speed (float, optional) : The requested speed to spin the spindle at.
            If ``speed`` is not specified the current interpreter speed setting
            (as set by the last S word) is used, taking into account the
            value of the spindle override if it is enabled.
        spindle (int, optional) : The number of the spindle to turn ON. If
            ``spindle`` is not specified spindle 0 is assumed.
    """
    if speed is None:
        speed = getSpeed()
    CMD.spindle(linuxcnc.SPINDLE_REVERSE, speed, spindle)

def _spindle_reverse_bindOk(speed=None, spindle=0, widget=None):
    if not _spindle_exists(spindle):
        return
    widget.setCheckable(True)
    STATUS.on.notify(lambda: _spindle_ok(spindle=spindle, widget=widget))
    STATUS.task_mode.onValueChanged(lambda: _spindle_ok(spindle=spindle, widget=widget))
    STATUS.spindle[spindle].direction.onValueChanged(lambda d: widget.setChecked(d == -1))

reverse.ok = _spindle_ok
reverse.bindOk = _spindle_reverse_bindOk


def off(spindle=0):
    """Turn a spindle OFF.

    ActionButton syntax to stop spindle 0 (spindle 0 is default)
    ::

        spindle.off

    ActionButton syntax to stop spindle 1
    ::

        spindle.1.off

    Args:
        spindle (int, optional) : The number of the spindle to turn OFF. If
            ``spindle`` is not specified spindle 0 is assumed.
    """
    CMD.spindle(linuxcnc.SPINDLE_OFF, spindle)

off.ok = _spindle_ok
off.bindOk = _spindle_bindOk


def faster(spindle=0):
    """Increase spindle speed by 100rpm.

    ActionButton syntax to increase spindle 0 speed (spindle 0 is default)
    ::

        spindle.faster

    ActionButton syntax to increase spindle 1
    ::

        spindle.1.faster

    Args:
        spindle (int, optional) : The number of the spindle to increase the
            speed of. If ``spindle`` is not specified spindle 0 is assumed.
    """
    CMD.spindle(linuxcnc.SPINDLE_INCREASE, spindle)

faster.ok = _spindle_ok
faster.bindOk = _spindle_bindOk


def slower(spindle=0):
    """Decrease spindle speed by 100rpm.

    ActionButton syntax to decrease spindle 0 speed (spindle 0 is default)
    ::

        spindle.slower

    ActionButton syntax to decrease spindle 1 speed
    ::

        spindle.1.slower


    Args:
        spindle (int, optional) : The number of the spindle to decrease the
            speed of. If ``spindle`` is not specified spindle 0 is assumed.
    """
    CMD.spindle(linuxcnc.SPINDLE_DECREASE, spindle)

slower.ok = _spindle_ok
slower.bindOk = _spindle_bindOk


# Spindle CONSTANT
def constant(spindle=0):
    """Unclear"""
    CMD.spindle(linuxcnc.SPINDLE_CONSTANT)

constant.ok = _spindle_ok
constant.bindOk = _spindle_bindOk


def getSpeed(spindle=0):
    """Gets the interpreter's speed setting for the specified spindle.

    Args:
        spindle (int, optional) : The number of the spindle to get the speed
            of. If ``spindle`` is not specified spindle 0 is assumed.

    Returns:
        float: The interpreter speed setting, with any override applied if
        override enabled.
    """
    raw_speed = STAT.settings[2]
    if raw_speed == 0:
        raw_speed = abs(DEFAULT_SPEED)

    return raw_speed

#==============================================================================
# Spindle Override
#==============================================================================

def override(override, spindle=0):
    """Set spindle override percentage.

    Args:
        override (float) : The desired spindle override in percent.
        spindle (float, optional) : The number of the spindle to apply the
            override to. If ``spindle`` is not specified spindle 0 is assumed.
    """
    CMD.spindleoverride(float(override) / 100, spindle)

def _or_reset(spindle=0):
    CMD.spindleoverride(1.0, spindle)

def _or_enable(spindle=0):
    CMD.set_spindle_override(1, spindle)

def _or_disable(spindle=0):
    CMD.set_spindle_override(0, spindle)

def _or_toggle_enable(spindle=0):
    if STAT.spindle[spindle]['override_enabled']:
        override.disable()
    else:
        override.enable()

override.reset = _or_reset
override.enable = _or_enable
override.disable = _or_disable
override.toggle_enable = _or_toggle_enable

def _or_ok(value=100, spindle=0, widget=None):
    if spindle not in SPINDLES:
            ok = False
            msg = "No spindle No. {}".format(spindle)
    elif STAT.task_state != linuxcnc.STATE_ON:
        ok = False
        msg = "Machine must be ON"
    elif STAT.spindle[0]['override_enabled'] != 1:
        ok = False
        msg = "Override not enabled"
    else:
        ok = True
        msg = ""

    _or_ok.msg = msg

    if widget is not None:
        widget.setEnabled(ok)
        widget.setStatusTip(msg)
        widget.setToolTip(msg)

    return ok

def _or_bindOk(value=100, spindle=0, widget=None):
    if not _spindle_exists(spindle):
        return

    # This will work for any widget
    STATUS.task_state.onValueChanged(lambda: _or_ok(widget=widget))
    STATUS.spindle[spindle].override_enabled.onValueChanged(lambda: _or_ok(widget=widget))

    try:
        # these will only work for QSlider or QSpinBox
        widget.setMinimum(INFO.minSpindleOverride() * 100)
        widget.setMaximum(INFO.maxSpindleOverride() * 100)

        try:
            widget.setSliderPosition(100)
            STATUS.spindle[spindle].override.onValueChanged(
                lambda v: widget.setSliderPosition(v * 100))

        except AttributeError:
            widget.setValue(100)
            STATUS.spindle[spindle].override.onValueChanged(
                lambda v: widget.setValue(v * 100))

        override(100)

    except AttributeError:
        pass
    except:
        LOG.exception('Error in spindle.override bindOk')

override.ok = override.reset.ok = _or_ok
override.bindOk = override.reset.bindOk = _or_bindOk

def _or_enable_ok(spindle=0, widget=None):
    if spindle not in SPINDLES:
        ok = False
        msg = "No spindle No. {}".format(spindle)
    elif STAT.task_state != linuxcnc.STATE_ON:
        ok = False
        msg = "Machine must be ON"
    elif STAT.interp_state != linuxcnc.INTERP_IDLE:
        ok = False
        msg = "Machine must be IDLE"
    else:
        ok = True
        msg = ""

    _or_enable_ok.msg = msg

    if widget is not None:
        widget.setEnabled(ok)
        widget.setStatusTip(msg)
        widget.setToolTip(msg)

    return ok

def _or_enable_bindOk(spindle=0, widget=None):
    if not _spindle_exists(spindle):
        return

    STATUS.task_state.onValueChanged(lambda: _or_enable_ok(spindle, widget))
    STATUS.interp_state.onValueChanged(lambda: _or_enable_ok(spindle, widget))
    STATUS.spindle[spindle].override_enabled.onValueChanged(widget.setChecked)

override.enable.ok = override.disable.ok = override.toggle_enable.ok = _or_enable_ok
override.enable.bindOk = override.disable.bindOk = override.toggle_enable.bindOk = _or_enable_bindOk

#==============================================================================
# Spindle Brake
#==============================================================================

class brake:
    """Spindle brake Group"""
    @staticmethod
    def on(spindle=0):
        """Set spindle brake ON.

        ActionButton syntax to engage spindle 0 brake (spindle 0 is default)
        ::

            spindle.brake.on

        ActionButton syntax to engage spindle 1 brake
        ::

            spindle.1.brake.on

        Args:
            spindle (int, optional) : The number of the spindle to apply the
                override to. If ``spindle`` is not specified spindle 0 is assumed.
        """
        CMD.brake(linuxcnc.BRAKE_ENGAGE, spindle)

    @staticmethod
    def off(spindle=0):
        """Set spindle brake OFF.

        ActionButton syntax to disengage spindle 0 brake (spindle 0 is default)
        ::

            spindle.brake.off

        ActionButton syntax to disengage spindle 1 brake
        ::

            spindle.1.brake.off


        Args:
            spindle (int, optional) : The number of the spindle to apply the
                override to. If ``spindle`` is not specified spindle 0 is assumed.
        """
        CMD.brake(linuxcnc.BRAKE_RELEASE, spindle)

    @staticmethod
    def toggle(spindle=0):
        """Toggle spindle brake ON/OFF.

        ActionButton syntax to toggle spindle 0 brake (spindle 0 is default)
        ::

            spindle.brake.toggle

        ActionButton syntax to toggle spindle 1 brake
        ::

            spindle.1.brake.toggle

        Args:
            spindle (int, optional) : The number of the spindle to apply the
                override to. If ``spindle`` is not specified spindle 0 is assumed.
        """
        if brake.is_on():
            brake.off()
        else:
            brake.on()

    @staticmethod
    def is_on(spindle=0):
        return STAT.spindle[spindle]['brake'] == linuxcnc.BRAKE_ENGAGE

def _brake_is_on(spindle=0):
    return STAT.spindle[spindle]['brake'] == linuxcnc.BRAKE_ENGAGE

def _brake_bind_ok(spindle=0, widget=None):
    if not _spindle_exists(spindle):
        return

    STATUS.on.notify(lambda: _spindle_ok(spindle=spindle, widget=widget))
    STATUS.task_mode.onValueChanged(lambda: _spindle_ok(spindle=spindle, widget=widget))
    STATUS.spindle[spindle].brake.onValueChanged(widget.setChecked)

brake.on.ok = brake.off.ok = brake.toggle.ok = _spindle_ok
brake.on.bindOk = brake.off.bindOk = brake.toggle.bindOk = _brake_bind_ok





import os

from qtpy import uic
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog

from qtpyvcp.utilities.logger import getLogger

LOG = getLogger(__name__)

class BaseDialog(QDialog):
    """Base Dialog

    Base QtPyVCP dialog class.

    This is intended to be used as a base class for custom dialogs, as well
    as a provider for use in YAML config files. This allows loading custom
    dialogs from .ui files without needing to write any python code.

    You can launch dialogs using a 
    :doc:`Dialog Button </widgets/buttons/index>` or
    from a window menu item.

    Example:

        YAML config for loading a custom dialog called `my_dialog` from a .ui
        file named ``my_diloag.ui`` located in the same dir as the .yml file::

            dialogs:
              my_dialog:
                provider: qtpyvcp.widgets.dialogs.base_dialog:BaseDialog
                kwargs:
                  ui_file: {{ file.dir }}/my_dialog.ui
                  title: My Dialog Title    # optional, set the dialog title
                  modal: false              # optional, whether the dialog is modal
                  popup: false              # optional, whether the dialog is a popup
                  frameless: false          # optional, whether the dialog is frameless
                  stay_on_top: true         # optional, whether the dialog stays on top

    Args:
        parent (QWidget, optional) : The dialog's parent window, or None.
        ui_file (str, optional) : The path of a .ui file to load the dialog
            from. The ui base widget should be a QDialog.
        title (str, optional) : The title to use for the dialog. This will
            override any title property set in QtDesigner.
        modal (bool, optional) : Whether the dialog should be application modal.
            This will override any modality hints set in QtDesigner.
        frameless (bool, optional) : Whether the window has a frame or not.
            If the window does not have a frame you will need some way to
            close it, like an Ok or Cancel button.
        popup: (bool, optional) : Makes the dialog use a frame less window
            that automatically hides when it looses focus.
        stay_on_top (bool, optional) : Sets the stay on top hint window flag.
            This overrides any window flags set in QtDesiger.
    """
    def __init__(self, parent=None, ui_file=None, title=None, modal=None,
                 popup=None, frameless=None, stay_on_top=None):
        super(BaseDialog, self).__init__(parent)

        if ui_file is not None:
            self.loadUiFile(ui_file)

        if title is not None:
            self.setWindowTitle(title)

        if modal is not None:
            if modal:
                self.setWindowModality(Qt.ApplicationModal)
            else:
                self.setWindowModality(Qt.NonModal)

        if popup is not None:
            self.setWindowFlags(Qt.Popup)

        if frameless is not None:
            self.setWindowFlag(Qt.FramelessWindowHint, frameless)

        if stay_on_top is not None:
            self.setWindowFlag(Qt.WindowStaysOnTopHint, stay_on_top)

    def loadUiFile(self, ui_file):
        """Load dialog from a .ui file.

        The .ui file base class should be a QDialog.

        Args:
            ui_file (str) : path to the .ui file to load.
        """
        ui_file = os.path.realpath(ui_file)
        if not os.path.isfile(ui_file):
            LOG.error("Specified UI for dialog does not exist: %s", ui_file)
            return

        LOG.debug("Loading dialog from ui_file: %s", ui_file)
        uic.loadUi(ui_file, self)

    def setWindowFlag(self, flag, on):
        """BackPort QWidget.setWindowFlag() implementation from Qt 5.9

        This method was introduced in Qt 5.9 so is not present
        in Qt 5.7.1 which is standard on Debian 9 (stretch), so
        add our own implementation.
        """
        if on:
            # add flag
            self.setWindowFlags(self.windowFlags() | flag)
        else:
            # remove flag
            self.setWindowFlags(self.windowFlags() ^ flag)



#import sys;sys.path.append(r'/root/.p2/pool/plugins/org.python.pydev.core_7.5.0.202001101138/pysrc')
#import pydevd;pydevd.settrace()

import os
from collections import OrderedDict

from qtpyvcp.lib.types import DotDict

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

QTPYVCP_DIR = os.path.abspath(os.path.dirname(__file__))
TOP_DIR = os.path.dirname(QTPYVCP_DIR)

DEFAULT_CONFIG_FILE = os.path.join(QTPYVCP_DIR, 'yaml_lib/default_config.yml')

# globals
CONFIG = {}
OPTIONS = DotDict()
PLUGINS = OrderedDict()  # Ordered dict so can terminate in order initiated
DIALOGS = {}
WINDOWS = {}
SETTINGS = {}

# for backward compatibility
from qtpyvcp.app import run
def run_vcp(*args, **kwargs):
    run(*args, **kwargs)



#!/usr/bin/env python

"""Main entry point for MiniVCP.

This module contains the code required to run MiniVCP directly, without
having to pass it as an argument to qtpyvcp. It handles parsing command
line args and starting the main application.

Example:
    Assuming the dir this file is located in is on the PATH, you can
    launch MiniVCP by saying::

        $ mini --ini=/path/to/config.ini [options ...]

    Run with the --help option to print a full list of options.

"""

__version__ = '0.0.1'

import os
import qtpyvcp

VCP_DIR = os.path.realpath(os.path.dirname(__file__))
VCP_CONFIG_FILE = os.path.join(VCP_DIR, 'mini.yml')


def main(opts=None):

    if opts is None:
        from qtpyvcp.utilities.opt_parser import parse_opts
        opts = parse_opts(vcp_cmd='mini',
                          vcp_name='Mini',
                          vcp_version=__version__)

    # run the main application
    qtpyvcp.app.run(opts, VCP_CONFIG_FILE)


if __name__ == '__main__':
    main()



from qtpy.QtWidgets import QCheckBox
from qtpy.QtCore import Property

from qtpyvcp.widgets import CMDWidget
from qtpyvcp.actions import bindWidget

class ActionCheckBox(QCheckBox, CMDWidget):
    """General purpose checkbox for triggering QtPyVCP actions.

    Args:
        parent (QWidget, optional) : The parent widget of the checkbox, or None.
        action (str, optional) : The name of the action the checkbox should trigger.
    """
    def __init__(self, parent=None, action=None):
        super(ActionCheckBox, self).__init__(parent)

        self._action_name = ''
        if action is not None:
            self.actionName = action

    @Property(str)
    def actionName(self):
        """Property for the name of the action the checkbox triggers (str).

        When this property is set it calls :meth:`QtPyVCP.actions.bindWidget`
        to bind the widget to the action.
        """
        return self._action_name

    @actionName.setter
    def actionName(self, action_name):
        self._action_name = action_name
        bindWidget(self, action_name)



from qtpyvcp.widgets.qtdesigner import _DesignerPlugin, _PluginExtension

from action_button import ActionButton
class ActionButtonPlugin(_DesignerPlugin):
    def pluginClass(self):
        return ActionButton

from action_checkbox import ActionCheckBox
class ActionCheckBoxPlugin(_DesignerPlugin):
    def pluginClass(self):
        return ActionCheckBox

from action_spinbox import ActionSpinBox
class ActionSpinBoxPlugin(_DesignerPlugin):
    def pluginClass(self):
        return ActionSpinBox

from subcall_button import SubCallButton
class MacroButtonPlugin(_DesignerPlugin):
    def pluginClass(self):
        return SubCallButton
    def toolTip(self):
        return "Execute a macro"

from led_button import LEDButton
class LedButtonPlugin(_DesignerPlugin):
    def pluginClass(self):
        return LEDButton
    def toolTip(self):
        return "LED button"

from mdi_button import MDIButton
class MDIButtonPlugin(_DesignerPlugin):
    def pluginClass(self):
        return MDIButton

from dialog_button import DialogButton
class DialogButtonPlugin(_DesignerPlugin):
    def pluginClass(self):
        return DialogButton


"""Main QtPyVCP Application Module

Contains the VCPApplication class with core function and VCP loading logic.
"""
import os
import sys
import imp
import inspect
from pkg_resources import iter_entry_points

from qtpy import API
from qtpy.QtGui import QFontDatabase
from qtpy.QtCore import QTimer, Slot, Qt
from qtpy.QtWidgets import QApplication, QStyleFactory

import qtpyvcp

from qtpyvcp.utilities.logger import initBaseLogger
from qtpyvcp.plugins import getPlugin
from qtpyvcp.widgets.base_widgets.base_widget import VCPPrimitiveWidget
from qtpyvcp.widgets.form_widgets.main_window import VCPMainWindow

# initialize logging. If a base logger was already initialized in a startup
# script (e.g. vcp_launcher.py), then that logger will be returned, otherwise
# this will initialise a base logger with default log level of DEBUG
LOG = initBaseLogger('qtpyvcp')

# Needed to silence this PySide2 warning:
#    Qt WebEngine seems to be initialized from a plugin. Please set
#    Qt::AA_ShareOpenGLContexts using QCoreApplication::setAttribute
#    before constructing QGuiApplication.
if API == 'pyside2':
    from qtpy.QtCore import Qt
    QApplication.setAttribute(Qt.AA_ShareOpenGLContexts)


class VCPApplication(QApplication):

    def __init__(self, theme=None, stylesheet=None, custom_fonts=[]):
        app_args = (qtpyvcp.OPTIONS.command_line_args or "").split()
        super(VCPApplication, self).__init__(app_args)

        opts = qtpyvcp.OPTIONS

        self.status = getPlugin('status')

        self.initialiseDataPlugins()

        theme = opts.theme or theme
        if theme is not None:
            self.setStyle(QStyleFactory.create(theme))

        stylesheet = opts.stylesheet or stylesheet
        if stylesheet is not None:
            self.loadStylesheet(stylesheet, opts.develop)

        if custom_fonts:
            for font in custom_fonts:
                self.loadFont(font)

        # self.window = self.loadVCPMainWindow(opts, vcp_file)
        # if self.window is not None:
        #     self.window.show()

        if opts.hide_cursor:
            from qtpy.QtGui import QCursor
            self.setOverrideCursor(QCursor(Qt.BlankCursor))

        # Performance monitoring
        if opts.perfmon:
            import psutil
            self.perf = psutil.Process()
            self.perf_timer = QTimer()
            self.perf_timer.setInterval(2000)
            self.perf_timer.timeout.connect(self.logPerformance)
            self.perf_timer.start()

        self.aboutToQuit.connect(self.terminate)

    def loadVCPMainWindow(self, opts, vcp_file=None):
        """
        Loads a VCPMainWindow instance defined by a Qt .ui file, a Python .py
        file, or from a VCP python package.

        Parameters
        ----------
        vcp_file : str
            The path or name of the VCP to load.
        opts : OptDict
            A OptDict of options to pass to the VCPMainWindow subclass.

        Returns
        -------
        VCPMainWindow instance
        """
        vcp = opts.vcp or vcp_file
        if vcp is None:
            return

        if os.path.exists(vcp):

            vcp_path = os.path.realpath(vcp)
            if os.path.isfile(vcp_path):
                directory, filename = os.path.split(vcp_path)
                name, ext = os.path.splitext(filename)
                if ext == '.ui':
                    LOG.info("Loading VCP from UI file: yellow<{}>".format(vcp))
                    return VCPMainWindow(opts=opts, ui_file=vcp_path)
                elif ext == '.py':
                    LOG.info("Loading VCP from PY file: yellow<{}>".format(vcp))
                    return self.loadPyFile(vcp_path, opts)
            elif os.path.isdir(vcp_path):
                LOG.info("VCP is a directory")
                # TODO: Load from a directory if it has a __main__.py entry point
        else:
            try:
                entry_points = {}
                for entry_point in iter_entry_points(group='qtpyvcp.example_vcp'):
                    entry_points[entry_point.name] = entry_point
                for entry_point in iter_entry_points(group='qtpyvcp.vcp'):
                    entry_points[entry_point.name] = entry_point
                window = entry_points[vcp.lower()].load()
                return window(opts=opts)
            except:
                LOG.exception("Failed to load entry point")

        LOG.critical("VCP could not be loaded: yellow<{}>".format(vcp))
        sys.exit()

    def loadPyFile(self, pyfile, opts):
        """
        Load a .py file, performs some sanity checks to try and determine
        if the file actually contains a valid VCPMainWindow subclass, and if
        the checks pass, create and return an instance.

        This is an internal method, users will usually want to use `loadVCP` instead.

        Parameters
        ----------
        pyfile : str
            The path to a .py file to load.
        opts : OptDict
            A OptDict of options to pass to the VCPMainWindow subclass.

        Returns
        -------
        VCPMainWindow instance
        """
        # Add the pyfile module directory to the python path, so that submodules can be loaded
        module_dir = os.path.dirname(os.path.abspath(pyfile))
        sys.path.append(module_dir)

        # Load the module. It's attributes can be accessed via `python_vcp.attr`
        module = imp.load_source('python_vcp', pyfile)

        classes = [obj for name, obj in inspect.getmembers(module)
                   if inspect.isclass(obj)
                   and issubclass(obj, VCPMainWindow)
                   and obj != VCPMainWindow]
        if len(classes) == 0:
            raise ValueError("Invalid File Format."
                             " {} has no class inheriting from VCPMainWindow.".format(pyfile))
        if len(classes) > 1:
            LOG.warn("More than one VCPMainWindow class in file yellow<{}>."
                     " The first occurrence (in alphabetical order) will be used: {}"
            .format(pyfile, classes[0].__name__))
        cls = classes[0]

        # initialize and return the VCPMainWindow subclass
        return cls(opts=opts)

    def loadStylesheet(self, stylesheet, watch=False):
        """Loads a QSS stylesheet file containing styles to be applied
        to specific Qt and/or QtPyVCP widget classes.

        Args:
            stylesheet (str) : Path to the .qss stylesheet to load.
            watch (bool) : Whether to watch and re-load on .qss file changes.
        """

        def load(path):
            LOG.info("Loading global stylesheet: yellow<{}>".format(stylesheet))
            self.setStyleSheet("file:///" + path)

            if watch:
                from qtpy.QtCore import QFileSystemWatcher
                self.qss_file_watcher = QFileSystemWatcher()
                self.qss_file_watcher.addPath(stylesheet)
                self.qss_file_watcher.fileChanged.connect(load)

        load(stylesheet)

    def loadFont(self, font_path):
        """Loads a font file into the font database. The path can specify the
        location of a font file or a qresource."""
        LOG.debug("Loading custom font: %s" % font_path)
        res = QFontDatabase.addApplicationFont(font_path)
        if res != 0:
            LOG.error("Failed to load font: %s", font_path)

    def getWidget(self, name):
        """Searches for a widget by name in the application windows.

        Args:
            name (str) : ObjectName of the widget.

        Returns: QWidget
        """
        for win_name, obj in qtpyvcp.WINDOWS.items():
            if hasattr(obj, name):
                return getattr(obj, name)

        raise AttributeError("Could not find widget with name: %s" % name)

    @Slot()
    def logPerformance(self):
        """Logs total CPU usage (in percent), as well as per-thread usage.
        """
        with self.perf.oneshot():
            total_percent = self.perf.cpu_percent(interval=None)
            total_time = sum(self.perf.cpu_times())
            usage = ["{:.3f}".format(total_percent * ((t.system_time + t.user_time) / total_time)) for t in self.perf.threads()]

        LOG.info("Performance:\n"
                 "    Total CPU usage (%): {}\n"
                 "    Per Thread: {}\n"
                 .format(total_percent, ' '.join(usage)))

    def terminate(self):
        self.terminateWidgets()
        self.terminateDataPlugins()

    def initialiseWidgets(self):
        for w in self.allWidgets():
            if isinstance(w, VCPPrimitiveWidget):
                w.initialize()

    def terminateWidgets(self):
        LOG.debug("Terminating widgets")
        for w in self.allWidgets():
            if isinstance(w, VCPPrimitiveWidget):
                try:
                    w.terminate()
                except Exception:
                    LOG.exception('Error terminating %s widget', w)

    def initialiseDataPlugins(self):
        for plugin, obj in qtpyvcp.PLUGINS.items():
            LOG.debug("Initializing %s plugin", plugin)
            obj.initialise()

    def terminateDataPlugins(self):
        # terminate in reverse order, this is to prevent problems
        # when terminating plugins that used other plugins.
        for plugin, obj in reversed(qtpyvcp.PLUGINS.items()):
            LOG.debug("Terminating %s plugin", plugin)
            try:
                # try so that other plugins are terminated properly
                # even if one of them fails.
                obj.terminate()
            except Exception:
                LOG.exception('Error terminating %s plugin', plugin)



# Copyright (c) 2017-2018, SLAC National Accelerator Laboratory

# This file has been adapted from PyDM, and can be redistributed and/or
# modified in accordance with terms in conditions set forth in the BSD
# 3-Clause License. You can find the complete licence text in the LICENCES
# directory.

# Links:
#   PyDM Project: https://github.com/slaclab/pydm
#   PyDM Licence: https://github.com/slaclab/pydm/blob/master/LICENSE.md

from qtpy.QtWidgets import QAction
from qtpy.QtDesigner import QExtensionFactory, QPyDesignerTaskMenuExtension

Q_TYPEID = {
    'QDesignerTaskMenuExtension':      'org.qt-project.Qt.Designer.TaskMenu',
    'QDesignerContainerExtension':     'org.qt-project.Qt.Designer.Container',
    'QDesignerPropertySheetExtension': 'org.qt-project.Qt.Designer.PropertySheet'
}

class ExtensionFactory(QExtensionFactory):
    def __init__(self, parent=None):
        super(ExtensionFactory, self).__init__(parent)

    def createExtension(self, obj, iid, parent):
        if iid == Q_TYPEID['QDesignerTaskMenuExtension']:
            return TaskMenuExtension(obj, parent)
        # In the future we can expand to the others such as Property and etc
        # When the time comes...  we will need a new Extension and
        # the equivalent for TaskMenuExtension classes for the
        # property editor and an elif statement in here to instantiate it...


class TaskMenuExtension(QPyDesignerTaskMenuExtension):
    def __init__(self, widget, parent):
        super(TaskMenuExtension, self).__init__(parent)

        self.widget = widget
        self.__actions = None
        self.__extensions = []
        extensions = getattr(widget, 'extensions', None)

        if extensions is not None:
            for ex in extensions:
                extension = ex(self.widget)
                self.__extensions.append(extension)

    def taskActions(self):
        if self.__actions is None:
            self.__actions = []
            for ex in self.__extensions:
                self.__actions.extend(ex.actions())

        return self.__actions

    def preferredEditAction(self):
        if self.__actions is None:
            self.taskActions()
        if self.__actions:
            return self.__actions[0]


class _PluginExtension(object):
    def __init__(self, widget):
        self.widget = widget
        self._actions = []

    def addTaskMenuAction(self, action_name, callback):
        action = QAction(action_name, self.widget)
        action.triggered.connect(callback)
        self._actions.append(action)

    def actions(self):
        return self._actions



import os

from qtpy.QtWidgets import QComboBox

from qtpyvcp.actions.program_actions import load as loadProgram
from qtpyvcp.plugins import getPlugin
from qtpyvcp.widgets.dialogs import getDialog


class RecentFileComboBox(QComboBox):
    def __init__(self, parent=None):
        super(RecentFileComboBox, self).__init__(parent)

        self.status = getPlugin('status')

        self.activated.connect(self.onItemActivated)
        self.updateRecentFiles(self.status.recent_files)

        self.insertItem(0, 'No File Loaded', None)
        self.setCurrentIndex(0)
        self.status.recent_files.notify(self.updateRecentFiles)

    def updateRecentFiles(self, recent_files):
        self.clear()
        for file in recent_files:
            self.addItem(os.path.basename(file), file)

        # Add separator and item to launch the file dialog
        self.insertSeparator(len(self.status.recent_files.getValue()))
        self.addItem("Browse for files ...", 'browse_files')

    def onItemActivated(self):
        data = self.currentData()
        if data == 'browse_files':
            getDialog('open_file').show()
        elif data is None:
            pass
        else:
            loadProgram(data)



import os
import linuxcnc

from qtpy.QtCore import QTimer, QFileSystemWatcher

from qtpyvcp.utilities.logger import getLogger
from qtpyvcp.app.runtime_config import RuntimeConfig
from qtpyvcp.plugins import DataPlugin, DataChannel

from qtpyvcp.utilities.info import Info

INFO = Info()
LOG = getLogger(__name__)
STAT = linuxcnc.stat()
CMD = linuxcnc.command()


class Status(DataPlugin):

    stat = STAT

    def __init__(self, cycle_time=100):
        super(Status, self).__init__()

        self.no_force_homing = INFO.noForceHoming()

        self.file_watcher = None

        # recent files
        self.max_recent_files = 10
        with RuntimeConfig('~/.axis_preferences') as rc:
            files = rc.get('DEFAULT', 'recentfiles', default=[])
        files = [file for file in files if os.path.exists(file)]
        self.recent_files.setValue(files)

        # MDI history
        self._max_mdi_history_length = 100
        self._mdi_history_file = INFO.getMDIHistoryFile()
        self.loadMdiHistory(self._mdi_history_file)

        self.jog_increment = 0  # jog
        self.step_jog_increment = INFO.getIncrements()[0]
        self.jog_mode = True
        self.linear_jog_velocity = INFO.getJogVelocity()
        self.angular_jog_velocity = INFO.getJogVelocity()

        try:
            STAT.poll()
        except:
            pass

        excluded_items = ['axis', 'joint', 'spindle', 'poll', 'command', 'debug']

        self.old = {}
        # initialize data channels
        for item in dir(STAT):
            if item in self.channels:
                self.old[item] = getattr(STAT, item)
                self.channels[item].setValue(getattr(STAT, item))
            elif item not in excluded_items and not item.startswith('_'):
                self.old[item] = getattr(STAT, item)
                chan = DataChannel(doc=item)
                chan.setValue(getattr(STAT, item))
                self.channels[item] = chan
                setattr(self, item, chan)

        # add joint status channels
        self.joint = tuple(JointStatus(jnum) for jnum in range(9))
        for joint in self.joint:
            for chan, obj in joint.channels.items():
                self.channels['joint.{}.{}'.format(joint.jnum, chan)] = obj

        # add spindle status channels
        self.spindle = tuple(SpindleStatus(snum) for snum in range(8))
        for spindle in self.spindle:
            for chan, obj in spindle.channels.items():
                self.channels['spindle.{}.{}'.format(spindle.snum, chan)] = obj

        self.all_axes_homed.value = False
        self.homed.notify(self.all_axes_homed.setValue)
        self.enabled.notify(self.all_axes_homed.setValue)

        # Set up the periodic update timer
        self.timer = QTimer()
        self._cycle_time = cycle_time
        self.timer.timeout.connect(self._periodic)

        self.on.settable = True
        self.task_state.notify(lambda ts:
                               self.on.setValue(ts == linuxcnc.STATE_ON))

    recent_files = DataChannel(doc='List of recently loaded files', settable=True, data=[])

    def loadMdiHistory(self, fname):
        """Load MDI history from file."""
        mdi_history = []
        if os.path.isfile(fname):
            with open(fname, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()
                    mdi_history.append(line)

        mdi_history.reverse()
        self.mdi_history.setValue(mdi_history)

    def saveMdiHistory(self, fname):
        """Write MDI history to file."""
        with open(fname, 'w') as fh:
            cmds = self.mdi_history.value
            cmds.reverse()
            for cmd in cmds:
                fh.write(cmd + '\n')

    @DataChannel
    def axis_mask(self, chan, format='int'):
        """Axes as configured in the [TRAJ]COORDINATES INI option.

        To return the string in a status label::

            status:axis_mask
            status:axis_mask?string
            status:axis_mask?list

        :returns: the configured axes
        :rtype: int, list, str
        """

        if format == 'list':

            mask = '{0:09b}'.format(self.stat.axis_mask or 7)

            axis_list = []
            for anum, enabled in enumerate(mask[::-1]):
                if enabled == '1':
                    axis_list.append(anum)

            return axis_list

        return self.stat.axis_mask

    @axis_mask.tostring
    def axis_mask(self, chan):
        axes = ''
        for anum in self.axis_mask.getValue(format='list'):
            axes += 'XYZABCUVW'[anum]

        return axes

    @DataChannel
    def mdi_history(self, chan):
        """List of recently issued MDI commands.
            Commands are stored in reverse chronological order, with the
            newest command at the front of the list, and oldest at the end.
            When the list exceeds the length given by MAX_MDI_COMMANDS the
            oldest entries will be dropped.

            Duplicate commands will not be removed, so that MDI History
            can be replayed via the queue meachanisim from a point in
            the history forward. The most recently issued
            command will always be at the front of the list.
        """
        return chan.value

    @mdi_history.setter
    def mdi_history(self, chan, new_value):
        if isinstance(new_value, list):
            chan.value = new_value[:self._max_mdi_history_length]
        else:
            #cmd = str(new_value.strip())
            cmds = chan.value
            #if cmd in cmds:
            #    cmds.remove(cmd)

            cmds.insert(0, new_value)
            chan.value = cmds[:self._max_mdi_history_length]

        chan.signal.emit(chan.value)

    def mdi_remove_entry(self, mdi_index):
        """Remove the indicated cmd by index reference"""
        # TODO: This has some potential code redundancy. Follow above pattern
        chan = self.mdi_history
        cmds = chan.value
        del cmds[mdi_index]
        chan.signal.emit(cmds)

    @DataChannel
    def on(self, chan):
        """True if machine power is ON."""
        return STAT.task_state == linuxcnc.STATE_ON

    @DataChannel
    def file(self, chan):
        """Currently loaded file including path"""
        return chan.value or 'No file loaded'

    @file.setter
    def file(self, chan, fname):
        if STAT.interp_state == linuxcnc.INTERP_IDLE \
                and STAT.call_level == 0:

            if self.file_watcher is not None:
                if self.file_watcher.files():
                    self.file_watcher.removePath(chan.value)
                if os.path.isfile(fname):
                    self.file_watcher.addPath(fname)

            chan.value = fname
            chan.signal.emit(fname)

    def updateFile(self, path):
        LOG.debug("Reloading edited G-Code file: %s", path)
        if os.path.isfile(path):
            self.file.signal.emit(path)
            CMD.program_open(path)

    @DataChannel
    def state(self, chan):
        """Current command execution status

        1) Done
        2) Exec
        3) Error

        To return the string in a status label::

            status:state?string

        :returns: current command execution state
        :rtype: int, str
        """
        return STAT.state

    @state.tostring
    def state(self, chan):
        states = {0: "N/A",
                       linuxcnc.RCS_DONE: "Done",
                       linuxcnc.RCS_EXEC: "Exec",
                       linuxcnc.RCS_ERROR: "Error"}

        return states[STAT.state]

    @DataChannel
    def exec_state(self, chan):
        """Current task execution state

        1) Error
        2) Done
        3) Waiting for Motion
        4) Waiting for Motion Queue
        5) Waiting for Pause
        6) --
        7) Waiting for Motion and IO
        8) Waiting for Delay
        9) Waiting for system CMD
        10) Waiting for spindle orient

        To return the string in a status label::

            status:exec_state?string

        :returns: current task execution error
        :rtype: int, str
        """
        return STAT.exec_state

    @exec_state.tostring
    def exec_state(self, chan):
        exec_states = {0: "N/A",
                        linuxcnc.EXEC_ERROR: "Error",
                        linuxcnc.EXEC_DONE: "Done",
                        linuxcnc.EXEC_WAITING_FOR_MOTION: "Waiting for Motion",
                        linuxcnc.EXEC_WAITING_FOR_MOTION_QUEUE: "Waiting for Motion Queue",
                        linuxcnc.EXEC_WAITING_FOR_IO: "Waiting for Pause",
                        linuxcnc.EXEC_WAITING_FOR_MOTION_AND_IO: "Waiting for Motion and IO",
                        linuxcnc.EXEC_WAITING_FOR_DELAY: "Waiting for Delay",
                        linuxcnc.EXEC_WAITING_FOR_SYSTEM_CMD: "Waiting for system CMD",
                        linuxcnc.EXEC_WAITING_FOR_SPINDLE_ORIENTED: "Waiting for spindle orient"}

        return exec_states[STAT.exec_state]

    @DataChannel
    def interp_state(self, chan):
        """Current state of RS274NGC interpreter

        1) Idle
        2) Reading
        3) Paused
        4) Waiting

        To return the string in a status label::

            status:interp_state?string

        :returns: RS274 interpreter state
        :rtype: int, str
        """
        return STAT.interp_state

    @interp_state.tostring
    def interp_state(self, chan):
        interp_states = {0: "N/A",
                            linuxcnc.INTERP_IDLE: "Idle",
                            linuxcnc.INTERP_READING: "Reading",
                            linuxcnc.INTERP_PAUSED: "Paused",
                            linuxcnc.INTERP_WAITING: "Waiting"}

        return interp_states[STAT.interp_state]


    @DataChannel
    def interpreter_errcode(self, chan):
        """Current RS274NGC interpreter return code

        0) Ok
        1) Exit
        2) Finished
        3) Endfile
        4) File not open
        5) Error

        To return the string in a status label::

            status:interpreter_errcode?string

        :returns: interp error code
        :rtype: int, str
        """
        return STAT.interpreter_errcode

    @interpreter_errcode.tostring
    def interpreter_errcode(self, chan):
        interpreter_errcodes = {0: "Ok",
                                1: "Exit",
                                2: "Finished",
                                3: "Endfile",
                                4: "File not open",
                                5: "Error"}

        return interpreter_errcodes[STAT.interpreter_errcode]

    @DataChannel
    def task_state(self, chan, query=None):
        """Current status of task

        1) E-Stop
        2) Reset
        3) Off
        4) On

        To return the string in a status label::

            status:task_state?string

        :returns: current task state
        :rtype: int, str
        """
        return STAT.task_state

    @task_state.tostring
    def task_state(self, chan):
        task_states = {0: "N/A",
                       linuxcnc.STATE_ESTOP: "E-Stop",
                       linuxcnc.STATE_ESTOP_RESET: "Reset",
                       linuxcnc.STATE_ON: "On",
                       linuxcnc.STATE_OFF: "Off"}

        return task_states[STAT.task_state]

    @DataChannel
    def task_mode(self, chan):
        """Current task mode

        1) Manual
        2) Auto
        3) MDI

        To return the string in a status label::

            status:task_mode?string

        :returns: current task mode
        :rtype: int, str
        """
        return STAT.task_mode

    @task_mode.tostring
    def task_mode(self, chan):
        task_modes = {0: "N/A",
                       linuxcnc.MODE_MANUAL: "Manual",
                       linuxcnc.MODE_AUTO: "Auto",
                       linuxcnc.MODE_MDI: "MDI"}

        return task_modes[STAT.task_mode]

    @DataChannel
    def motion_mode(self, chan):
        """Current motion controller mode

        1) Free
        2) Coord
        3) Teleop

        To return the string in a status label::

            status:motion_mode?string

        :returns: current motion mode
        :rtype: int, str
        """
        return STAT.motion_mode

    @motion_mode.tostring
    def motion_mode(self, chan):
        modes = {0: "N/A",
                  linuxcnc.TRAJ_MODE_COORD: "Coord",
                  linuxcnc.TRAJ_MODE_FREE: "Free",
                  linuxcnc.TRAJ_MODE_TELEOP: "Teleop"}

        return modes[STAT.motion_mode]

    @DataChannel
    def motion_type(self, chan, query=None):
        """Motion type

        0) None
        1) Traverse
        2) Linear Feed
        3) Arc Feed
        4) Tool Change
        5) Probing
        6) Rotary Index

        To return the string in a status label::

            status:motion_type?string

        :returns:  current motion type
        :rtype: int, str
        """
        return STAT.motion_type

    @motion_type.tostring
    def motion_type(self, chan):
        motion_types = {0: "None",
                        linuxcnc.MOTION_TYPE_TRAVERSE: "Traverse",
                        linuxcnc.MOTION_TYPE_FEED: "Linear Feed",
                        linuxcnc.MOTION_TYPE_ARC: "Arc Feed",
                        linuxcnc.MOTION_TYPE_TOOLCHANGE: "Tool Change",
                        linuxcnc.MOTION_TYPE_PROBING: "Probing",
                        linuxcnc.MOTION_TYPE_INDEXROTARY: "Rotary Index"}

        return motion_types[STAT.motion_type]

    @DataChannel
    def program_units(self, chan):
        """Program units

        Available as an integer, or in short or long string formats.

        1) in, Inches
        2) mm, Millimeters
        3) cm, Centimeters

        To return the string in a status label::

            status:program_units
            status:program_units?string
            status:program_units?string&format=long

        :returns: current program units
        :rtype: int, str
        """
        return STAT.program_units

    @program_units.tostring
    def program_units(self, chan, format='short'):
        if format == 'short':
            return ["N/A", "in", "mm", "cm"][STAT.program_units]
        else:
            return ["N/A", "Inches", "Millimeters", "Centimeters"][STAT.program_units]

    @DataChannel
    def linear_units(self, chan):
        """Machine linear units

        Available as float (units/mm), or in short or long string formats.

        To return the string in a status label::

            status:linear_units
            status:linear_units?string
            status:linear_units?string&format=long

        :returns: machine linear units
        :rtype: float, str
        """
        return STAT.linear_units

    @linear_units.tostring
    def linear_units(self, chan, format='short'):
        if format == 'short':
            return {0.0: "N/A", 1.0: "mm", 1 / 25.4: "in"}[STAT.linear_units]
        else:
            return {0.0: "N/A", 1.0: "Millimeters", 1 / 25.4: "Inches"}[STAT.linear_units]

    @DataChannel
    def gcodes(self, chan, fmt=None):
        """G-codes

        active G-codes for each modal group

        | syntax ``status:gcodes`` returns tuple of strings
        | syntax ``status:gcodes?raw`` returns tuple of integers
        | syntax ``status:gcodes?string`` returns str
        """
        if fmt == 'raw':
            return STAT.gcodes
        return chan.value

    @gcodes.tostring
    def gcodes(self, chan):
        return " ".join(chan.value)

    @gcodes.setter
    def gcodes(self, chan, gcodes):
        chan.value = tuple(["G%g" % (c/10.) for c in sorted(gcodes[1:]) if c != -1])
        chan.signal.emit(self.gcodes.value)

    @DataChannel
    def mcodes(self, chan, fmt=None):
        """M-codes

        active M-codes for each modal group

        | syntax ``status:mcodes`` returns tuple of strings
        | syntax ``status:mcodes?raw`` returns tuple of integers
        | syntax ``status:mcodes?string`` returns str
        """
        if fmt == 'raw':
            return STAT.mcodes
        return chan.value

    @mcodes.tostring
    def mcodes(self, chan):
        return " ".join(chan.value)

    @mcodes.setter
    def mcodes(self, chan, gcodes):
        chan.value = tuple(["M%g" % gcode for gcode in sorted(gcodes[1:]) if gcode != -1])
        chan.signal.emit(chan.value)

    @DataChannel
    def g5x_index(self, chan):
        """Current G5x work coord system

        | syntax ``status:g5x_index`` returns int
        | syntax ``status:g5x_index?string`` returns str
        """
        return STAT.g5x_index

    @g5x_index.tostring
    def g5x_index(self, chan):
        return ["G53", "G54", "G55", "G56", "G57", "G58",
                "G59", "G59.1", "G59.2", "G59.3"][STAT.g5x_index]

    @DataChannel
    def settings(self, chan, item=None):
        """Interpreter Settings

        Available Items:
            0) sequence_number
            1) feed
            2) speed

        :return: interpreter settings
        :rtype: tuple, int, float
        """
        if item is None:
            return STAT.settings
        return STAT.settings[{'sequence_number': 0, 'feed': 1, 'speed': 2}[item]]

    @DataChannel
    def homed(self, chan, anum=None):
        """Axis homed status

        If no axis number is specified returns a tuple of integers.
        If ``anum`` is specified returns True if the axis is homed, else False.

        Rules syntax::

            status:homed
            status:homed?anum=0

        Args:
         anum (int, optional) : the axis number to return the homed state of.

        :returns: axis homed states
        :rtype: tuple, bool

        """
        if anum is None:
            return STAT.homed
        return bool(STAT.homed[int(anum)])

    @DataChannel
    def all_axes_homed(self, chan):
        """All axes homed status

        True if all axes are homed or if [TRAJ]NO_FORCE_HOMING set in INI.

        If [TRAJ]NO_FORCE_HOMING is set in the INI the value will be come
        true as soon as the machine is turned on and the signal will be emitted,
        otherwise the signal will be emitted once all the axes defined in the
        INI have been homed.

        :returns: all homed
        :rtype: bool
        """
        return chan.value

    @all_axes_homed.setter
    def all_axes_homed(self, chan, homed):
        if self.no_force_homing:
            all_homed = True
        else:
            for anum in INFO.AXIS_NUMBER_LIST:
                if STAT.homed[anum] is not 1:
                    all_homed = False
                    break
            else:
                all_homed = True

        if all_homed != chan.value:
            chan.value = all_homed
            chan.signal.emit(chan.value)


    # this is used by File "qtpyvcp/qtpyvcp/actions/program_actions.py",
    # line 83, in _run_ok elif not STATUS.allHomed():

    def allHomed(self):
        if self.no_force_homing:
            return True
        for jnum in range(STAT.joints):
            if not STAT.joint[jnum]['homed']:
                return False
        return True

    def forceUpdateStaticChannelMembers(self):
        """Static items need a force update to operate properly with the
        gui rules.  This needs to be done with consideration to the
        data structure so as to not "break" things.
        """
        # TODO: add to this list as needed. Possible to externalise via yaml?
        self.old['axes'] = None

    def initialise(self):
        """Start the periodic update timer."""

        # watch the gcode file for changes and reload as needed
        self.file_watcher = QFileSystemWatcher()
        if self.file.value:
            self.file_watcher.addPath(self.file.value)
        self.file_watcher.fileChanged.connect(self.updateFile)

        LOG.debug("Starting periodic updates with %ims cycle time",
                  self._cycle_time)
        self.timer.start(self._cycle_time)

        self.forceUpdateStaticChannelMembers()

    def terminate(self):
        """Save persistent data on terminate."""

        # save recent files
        with RuntimeConfig('~/.axis_preferences') as rc:
            rc.set('DEFAULT', 'recentfiles', self.recent_files.value)

        # save MDI history
        self.saveMdiHistory(self._mdi_history_file)

    def _periodic(self):

        # s = time.time()

        try:
            STAT.poll()
        except Exception:
            LOG.warning("Status polling failed, is LinuxCNC running?", exc_info=True)
            self.timer.stop()
            return

        # status updates
        for item, old_val in self.old.iteritems():
            new_val = getattr(STAT, item)
            if new_val != old_val:
                self.old[item] = new_val
                self.channels[item].setValue(new_val)

        # joint status updates
        for joint in self.joint:
            joint._update()

        # spindle status updates
        for spindle in self.spindle:
            spindle._update()

        # print time.time() - s


class JointStatus(DataPlugin):
    def __init__(self, jnum):
        super(JointStatus, self).__init__()

        self.jnum = jnum
        self.jstat = STAT.joint[jnum]

        for key, value in self.jstat.items():
            chan = DataChannel(doc=key, data=value)
            self.channels[key] = chan
            setattr(self, key, chan)

    def _update(self):
        """Periodic joint item updates."""

        jstat = STAT.joint[self.jnum].items()
        changed_items = tuple(set(jstat) - set(self.jstat.items()))
        for item in changed_items:
            LOG.debug('JOINT_{0} {1}: {2}'.format(self.jnum, item[0], item[1]))
            self.channels[item[0]].setValue(item[1])

        self.jstat.update(jstat)


class SpindleStatus(DataPlugin):
    def __init__(self, snum):
        super(SpindleStatus, self).__init__()

        self.snum = snum
        self.sstat = STAT.spindle[snum]

        for key, value in self.sstat.items():
            chan = DataChannel(doc=key, data=value)
            self.channels[key] = chan
            setattr(self, key, chan)

    def _update(self):
        """Periodic spindle item updates."""

        sstat = STAT.spindle[self.snum].items()
        changed_items = tuple(set(sstat) - set(self.sstat.items()))
        for item in changed_items:
            LOG.debug('Spindle_{0} {1}: {2}'.format(self.snum, item[0], item[1]))
            self.channels[item[0]].setValue(item[1])

        self.sstat.update(sstat)
